from __future__ import annotations

from .shell_parser import ShellRedirection, extract_heredoc_bodies, parse_shell_script


def test_parse_shell_script_parses_command_lists_and_pipelines() -> None:
    program = parse_shell_script(
        script=(
            "cd /website && npm install tslib @types/react @types/react-dom "
            "--cache /agents/webmaster/.npm && npm run build && "
            "find dist -maxdepth 2 -type f | sort"
        )
    )

    assert len(program.items) == 4
    assert program.items[0].separator_before is None
    assert program.items[1].separator_before == "&&"
    assert program.items[2].separator_before == "&&"
    assert program.items[3].separator_before == "&&"
    assert program.items[0].pipeline.first_command.argv == ("cd", "/website")
    assert program.items[1].pipeline.first_command.argv == (
        "npm",
        "install",
        "tslib",
        "@types/react",
        "@types/react-dom",
        "--cache",
        "/agents/webmaster/.npm",
    )
    assert len(program.items[3].pipeline.commands) == 2
    assert program.items[3].pipeline.commands[0].argv == (
        "find",
        "dist",
        "-maxdepth",
        "2",
        "-type",
        "f",
    )
    assert program.items[3].pipeline.commands[1].argv == ("sort",)


def test_parse_shell_script_preserves_quoted_words() -> None:
    program = parse_shell_script(
        script=(
            "stat -c '%A %n' node_modules/rollup/dist/bin/rollup "
            "node_modules/.bin/rollup && "
            'sed -i \'s#"build": "rollup -c"#"build": '
            '"node ./node_modules/rollup/dist/bin/rollup -c"#\' package.json'
        )
    )

    assert len(program.items) == 2
    assert program.items[0].pipeline.first_command.argv == (
        "stat",
        "-c",
        "'%A %n'",
        "node_modules/rollup/dist/bin/rollup",
        "node_modules/.bin/rollup",
    )
    assert program.items[1].pipeline.first_command.argv == (
        "sed",
        "-i",
        '\'s#"build": "rollup -c"#"build": "node ./node_modules/rollup/dist/bin/rollup -c"#\'',
        "package.json",
    )


def test_parse_shell_script_strips_heredoc_bodies_and_keeps_redirections() -> None:
    program = parse_shell_script(
        script=(
            "cd /website && cat > public/index.html <<'EOF'\n"
            "<!doctype html>\n"
            "<html></html>\n"
            "EOF\n"
            "npm run build"
        )
    )

    assert len(program.items) == 3
    cat_command = program.items[1].pipeline.first_command
    assert cat_command.argv == ("cat",)
    assert cat_command.redirections == (
        ShellRedirection(operator=">", target="public/index.html"),
        ShellRedirection(operator="<<", target="'EOF'"),
    )
    assert program.items[2].pipeline.first_command.argv == ("npm", "run", "build")


def test_parse_shell_script_supports_dash_heredoc_operator() -> None:
    program = parse_shell_script(
        script=("python - <<-'PY'\nprint('hi')\nPY\necho done")
    )

    assert len(program.items) == 2
    command = program.items[0].pipeline.first_command
    assert command.argv == ("python", "-")
    assert command.redirections == (ShellRedirection(operator="<<-", target="'PY'"),)


def test_parse_shell_script_supports_file_descriptor_redirections() -> None:
    program = parse_shell_script(
        script=(
            "cd /data/pythontest && "
            "python3 -m pip install ruff >/tmp/ruff_install.log 2>&1 && "
            "python3 -m ruff check ."
        )
    )

    assert len(program.items) == 3
    install_command = program.items[1].pipeline.first_command
    assert install_command.argv == ("python3", "-m", "pip", "install", "ruff")
    assert install_command.redirections == (
        ShellRedirection(operator=">", target="/tmp/ruff_install.log"),
        ShellRedirection(operator="2>&", target="1"),
    )
    lint_command = program.items[2].pipeline.first_command
    assert lint_command.argv == ("python3", "-m", "ruff", "check", ".")


def test_extract_heredoc_bodies_returns_embedded_script_bodies_in_order() -> None:
    bodies = extract_heredoc_bodies(
        script=("python - <<'PY'\nprint('one')\nPY\ncat > out.txt <<'EOF'\nhello\nEOF")
    )

    assert bodies == ("print('one')", "hello")


def test_parse_shell_script_supports_if_shell_command() -> None:
    program = parse_shell_script(
        script="if test -f app.py; then echo ok; else echo nope; fi"
    )

    assert len(program.items) == 1
    command = program.items[0].pipeline.first_command
    assert command.executable == "if"
    assert "then" in command.text
    assert "else" in command.text
    assert "fi" in command.text


def test_parse_shell_script_supports_if_test_command_with_bang() -> None:
    program = parse_shell_script(
        script="if [ ! -f index.html ]; then cat > index.html <<'EOF'\nhi\nEOF\nfi"
    )

    assert len(program.items) == 1
    command = program.items[0].pipeline.first_command
    assert command.executable == "if"
    assert "[ ! -f index.html ]" in command.text
    assert command.compound is not None
    assert len(command.compound.programs) >= 1


def test_parse_shell_script_supports_for_shell_command() -> None:
    program = parse_shell_script(
        script="for file in README.md app.py; do echo $file; done"
    )

    assert len(program.items) == 1
    command = program.items[0].pipeline.first_command
    assert command.executable == "for"
    assert "in README.md app.py" in command.text
    assert "done" in command.text



from collections.abc import AsyncIterable
from typing import Optional
import json
from meshagent.api.room_server_client import (
    RoomException,
    RequiredToolkit,
    Requirement,
    RequiredSchema,
    RequiredTable,
)
from meshagent.api import (
    ToolContentSpec,
    ToolDescription,
    ToolkitDescription,
    StorageEntry,
)
from meshagent.tools import (
    Toolkit,
    FunctionTool,
    ToolContext,
    RemoteToolkit,
)

from meshagent.api.room_server_client import RoomClient
from .context import AgentSessionContext
import logging
import asyncio
import warnings

logger = logging.getLogger("agent")
_legacy_init_chat_context_warned: set[type] = set()


class AgentException(RoomException):
    pass


class RoomTool(FunctionTool):
    def __init__(
        self,
        *,
        toolkit_name: str,
        name,
        input_schema,
        output_spec: Optional[ToolContentSpec] = None,
        output_schema: Optional[dict] = None,
        title=None,
        description=None,
        rules=None,
        thumbnail_url=None,
        pricing: Optional[str] = None,
        participant_id: Optional[str] = None,
        on_behalf_of_id: Optional[str] = None,
        defs: Optional[dict] = None,
        strict: Optional[bool] = None,
    ):
        self._toolkit_name = toolkit_name
        self._participant_id = participant_id
        self._on_behalf_of_id = on_behalf_of_id
        if input_schema is None:
            input_schema = {
                "type": "object",
                "additionalProperties": True,
                "properties": {},
            }

        super().__init__(
            name=name,
            input_schema=input_schema,
            output_spec=output_spec,
            output_schema=output_schema,
            title=title,
            description=description,
            rules=rules,
            thumbnail_url=thumbnail_url,
            pricing=pricing,
            defs=defs,
            strict=True if strict is None else strict,
        )

    async def execute(self, context: ToolContext, **kwargs):
        result = await context.room.agents.invoke_tool(
            toolkit=self._toolkit_name,
            tool=self.name,
            participant_id=self._participant_id,
            on_behalf_of_id=self._on_behalf_of_id,
            input=kwargs,
            caller_context=context.caller_context,
        )

        if isinstance(result, AsyncIterable):
            raise RoomException(
                f"tool '{self._toolkit_name}.{self.name}' returned an iterable stream, which is not supported by RoomTool"
            )

        return result


class Agent:
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        requires: Optional[list[Requirement]] = None,
        annotations: Optional[list[str]] = None,
        skills_dirs: Optional[list[str]] = None,
    ):
        if name is not None:
            logger.warning(
                f"agent name property is deprecated and will be removed in a future version {name}"
            )

        if title is None:
            title = name
        self._title = title
        if description is None:
            description = ""

        self._description = description
        if requires is None:
            requires = []

        self._requires = requires

        if annotations is None:
            annotations = []

        self._annotations = annotations
        self._skills_dirs = None

    def get_requirements(self) -> list[Requirement]:
        return self._requires

    @property
    def description(self):
        return self._description

    @property
    def title(self):
        return self._title

    @property
    def requires(self):
        return self._requires

    @property
    def annotations(self):
        return self._annotations

    async def init_session(self) -> AgentSessionContext:
        legacy_initializer = type(self).init_chat_context
        if legacy_initializer is not Agent.init_chat_context:
            cls = type(self)
            if cls not in _legacy_init_chat_context_warned:
                warnings.warn(
                    (
                        f"{cls.__name__}.init_chat_context() is deprecated and will be removed in a future release. "
                        "Override init_session() instead."
                    ),
                    DeprecationWarning,
                    stacklevel=2,
                )
                _legacy_init_chat_context_warned.add(cls)
            return await legacy_initializer(self)
        return AgentSessionContext()

    # Backwards compatibility for existing subclasses overriding init_chat_context.
    async def init_chat_context(self) -> AgentSessionContext:
        warnings.warn(
            "init_chat_context() is deprecated and will be removed in a future release. Use init_session() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return AgentSessionContext()

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "requires": list(map(lambda x: x.to_json(), self.requires)),
            "annotations": self.annotations,
        }


async def install_required_table(*, room: RoomClient, table: RequiredTable):
    await room.database.create_table_with_schema(
        name=table.name,
        mode="create_if_not_exists",
        schema=table.schema,
        namespace=table.namespace,
    )

    indexes = await room.database.list_indexes(
        table=table.name, namespace=table.namespace
    )

    def index_exists(column: str):
        for i in indexes:
            if column in i.columns:
                return True

        return False

    for vi in table.vector_indexes or []:
        if not index_exists(vi):
            try:
                await room.database.create_vector_index(
                    table=table.name,
                    column=vi,
                    namespace=table.namespace,
                    replace=True,
                )
            except Exception as e:
                logger.warning(f"unable to create vector index {e}", exec_info=e)

    for ti in table.full_text_search_indexes or []:
        if not index_exists(ti):
            try:
                await room.database.create_full_text_search_index(
                    table=table.name,
                    column=ti,
                    namespace=table.namespace,
                    replace=True,
                )
            except Exception as e:
                logger.warning(
                    f"unable to create full text search index {e}",
                    exec_info=e,
                )

    for si in table.scalar_indexes or []:
        if not index_exists(si):
            try:
                await room.database.create_scalar_index(
                    table=table.name,
                    column=si,
                    namespace=table.namespace,
                    replace=True,
                )
            except Exception as e:
                logger.warning(f"unable to create scalar index {e}", exec_info=e)

    logger.info(f"optimizing table {table.name} in {table.namespace}")

    # TODO: use index_stats to determine when indexes need to be updated
    await room.database.optimize(table=table.name, namespace=table.namespace)


class SingleRoomAgent(Agent):
    def __init__(
        self,
        *,
        name=None,
        title=None,
        description=None,
        requires=None,
        annotations: Optional[list[str]] = None,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            annotations=annotations,
        )
        self._room = None
        self._exposed_toolkits = None

    async def get_exposed_toolkits(self) -> list[RemoteToolkit]:
        return []

    async def start(self, *, room: RoomClient) -> None:
        if self._room is not None:
            raise RoomException("agent is already started")

        self._room = room

        await self.install_requirements()

        self._exposed_toolkits = await self.get_exposed_toolkits()

        for tk in self._exposed_toolkits:
            await tk.start(room=room)

    async def stop(self) -> None:
        for tk in self._exposed_toolkits:
            await tk.stop()

        self._room = None

    @property
    def room(self) -> RoomClient:
        return self._room

    @property
    def name(self) -> str:
        return self._room.local_participant.get_attribute("name")

    async def install_requirements(self, participant_id: Optional[str] = None):
        schemas_by_name = dict[str, StorageEntry]()
        toolkits_by_name = dict[str, ToolkitDescription]()

        async def refresh_schemas():
            schemas = await self._room.storage.list(path=".schemas")

            for schema in schemas:
                schemas_by_name[schema.name] = schema

        async def refresh_tools():
            toolkits_by_name.clear()

            visible_tools = await self._room.agents.list_toolkits(
                participant_id=participant_id
            )
            for toolkit_description in visible_tools:
                toolkits_by_name[toolkit_description.name] = toolkit_description

        installed = False

        await refresh_tools()
        await refresh_schemas()

        builtin_agents_url = "http://localhost:8080"

        for requirement in self.get_requirements():
            if isinstance(requirement, RequiredToolkit):
                if requirement.name == "ui":
                    # TODO: maybe requirements can be marked as non installable?
                    continue

                if requirement.name not in toolkits_by_name:
                    if not requirement.callable:
                        if requirement.timeout == 0:
                            logger.info(
                                f"{self.name} not waiting for toolkit {requirement.name}"
                            )
                            continue

                        async with asyncio.timeout(requirement.timeout):
                            logger.info(
                                f"{self.name} waiting for toolkit {requirement.name}"
                            )

                            while requirement.name not in toolkits_by_name:
                                await refresh_tools()
                                await asyncio.sleep(1)

                    else:
                        installed = True
                        logger.info(
                            f"{self.name} calling required tool into room {requirement.name}"
                        )

                        if requirement.name.startswith(
                            "https://"
                        ) or requirement.name.startswith("http://"):
                            url = requirement.name
                        else:
                            url = f"{builtin_agents_url}/toolkits/{requirement.name}"

                        await self._room.agents.make_call(
                            url=url, name=requirement.name, arguments={}
                        )

            elif isinstance(requirement, RequiredSchema):
                if requirement.schema is not None:
                    logger.info(
                        f"{self.name} installing required schema {requirement.name} from json"
                    )
                    await self._room.storage.upload(
                        path=f".schemas/{requirement.name}.json",
                        overwrite=True,
                        data=json.dumps(requirement.schema.to_json()).encode(),
                    )

                elif requirement.name not in schemas_by_name:
                    installed = True

                    if not requirement.callable:
                        if requirement.timeout == 0:
                            logger.info(
                                f"{self.name} not waiting for schema {requirement.name}"
                            )
                            continue

                        async with asyncio.timeout(requirement.timeout):
                            logger.info(
                                f"{self.name} waiting for schema {requirement.name}"
                            )

                            while requirement.name not in schemas_by_name:
                                await refresh_schemas()
                                await asyncio.sleep(1)

                    else:
                        logger.info(
                            f"{self.name} installing required schema {requirement.name} from registry"
                        )

                        if requirement.name.startswith(
                            "https://"
                        ) or requirement.name.startswith("http://"):
                            url = requirement.name
                        else:
                            url = f"{builtin_agents_url}/schemas/{requirement.name}"

                        await self._room.agents.make_call(
                            url=url, name=requirement.name, arguments={}
                        )

            elif isinstance(requirement, RequiredTable):
                logger.info(
                    f"ensuring required table exists {requirement.name} in {requirement.namespace}"
                )

                await install_required_table(room=self.room, table=requirement)

            else:
                raise RoomException("unsupported requirement")

        if installed:
            await asyncio.sleep(5)

    async def get_required_toolkits(self, context: ToolContext) -> list[Toolkit]:
        tool_target = context.caller
        if context.on_behalf_of is not None:
            tool_target = context.on_behalf_of

        toolkits_by_name = dict[str, ToolkitDescription]()
        toolkits_by_participant = dict[str, list[ToolkitDescription]]()
        toolkits = list[Toolkit]()

        visible_tools = await self._room.agents.list_toolkits(
            participant_id=tool_target.id
        )

        for toolkit_description in visible_tools:
            toolkits_by_name[toolkit_description.name] = toolkit_description

        for required_toolkit in self.requires:
            if isinstance(required_toolkit, RequiredToolkit):
                toolkit = None
                if required_toolkit.participant_name is None:
                    toolkit = toolkits_by_name.get(required_toolkit.name, None)
                else:
                    if required_toolkit.participant_name not in toolkits_by_participant:
                        toolkits_by_participant[
                            required_toolkit.participant_name
                        ] = await self._room.agents.list_toolkits(
                            participant_name=required_toolkit.participant_name
                        )

                    for tk in toolkits_by_participant[
                        required_toolkit.participant_name
                    ]:
                        if tk.name == required_toolkit.name:
                            toolkit = tk
                            break

                if toolkit is None:
                    if context.on_behalf_of is not None:
                        raise RoomException(
                            f"unable to get toolkit {required_toolkit.name} on behalf of {context.on_behalf_of}"
                        )
                    else:
                        raise RoomException(
                            f"unable to get toolkit {required_toolkit.name} for caller {context.caller.id}"
                        )

                room_tools = list[RoomTool]()

                if required_toolkit.tools is None:
                    for tool_description in toolkit.tools:
                        tool = RoomTool(
                            on_behalf_of_id=tool_target.id,
                            toolkit_name=toolkit.name,
                            name=tool_description.name,
                            description=tool_description.description,
                            input_schema=tool_description.input_schema,
                            output_spec=tool_description.output_spec,
                            output_schema=tool_description.output_schema,
                            title=tool_description.title,
                            thumbnail_url=tool_description.thumbnail_url,
                            participant_id=tool_target.id,
                            defs=tool_description.defs,
                            pricing=tool_description.pricing,
                            strict=tool_description.strict,
                        )
                        room_tools.append(tool)

                else:
                    tools_by_name = dict[str, ToolDescription]()
                    for tool_description in toolkit.tools:
                        tools_by_name[tool_description.name] = tool_description

                    for required_tool in required_toolkit.tools:
                        tool_description = tools_by_name.get(required_tool, None)
                        if tool_description is None:
                            raise RoomException(
                                f"unable to locate required tool {required_tool} in toolkit {required_toolkit.name}"
                            )

                        tool = RoomTool(
                            on_behalf_of_id=tool_target.id,
                            toolkit_name=toolkit.name,
                            name=tool_description.name,
                            description=tool_description.description,
                            input_schema=tool_description.input_schema,
                            output_spec=tool_description.output_spec,
                            output_schema=tool_description.output_schema,
                            title=tool_description.title,
                            thumbnail_url=tool_description.thumbnail_url,
                            participant_id=tool_description.participant_id,
                            defs=tool_description.defs,
                            pricing=tool_description.pricing,
                            strict=tool_description.strict,
                        )
                        room_tools.append(tool)

                toolkits.append(
                    Toolkit(
                        name=toolkit.name,
                        title=toolkit.title,
                        description=toolkit.description,
                        thumbnail_url=toolkit.thumbnail_url,
                        tools=room_tools,
                    )
                )

        return toolkits



from meshagent.agents.worker import Worker
from meshagent.tools import (
    RemoteToolkit,
    ToolContext,
    FunctionTool,
    Toolkit,
    FileContent,
)
from meshagent.tools.storage import StorageToolkit
from meshagent.api.room_server_client import TextDataType, RoomException
from email import message_from_bytes
from email.message import EmailMessage
from meshagent.api import RoomClient
from meshagent.api import RequiredTable
from email.policy import default
from meshagent.agents import AgentSessionContext
from datetime import datetime, timezone
import base64
import secrets

from typing import Literal, Optional
import json

import logging

import os

from pathlib import Path
from . import mail_common
from meshagent.agents.skills import to_prompt

logger = logging.getLogger("mail")


class _MissingAioSmtplib:
    async def send(self, *args: object, **kwargs: object) -> None:
        del args
        del kwargs
        raise ModuleNotFoundError(
            "aiosmtplib is required to use mail agent SMTP features"
        )


try:
    import aiosmtplib
except ModuleNotFoundError:
    aiosmtplib = _MissingAioSmtplib()

type MessageRole = Literal["user", "agent"]


class MailThreadContext:
    def __init__(self, *, chat: AgentSessionContext, message: dict, thread: list[dict]):
        self.chat = chat
        self.message = message
        self.thread = thread


SmtpConfiguration = mail_common.SmtpConfiguration


class NewEmailThreadWithAttachments(FunctionTool):
    def __init__(self, *, agent: "MailBot"):
        self.agent = agent
        super().__init__(
            name="new_email_thread",
            title="New Email Thread",
            description="Starts a new email thread that is managed by the mailbot",
            input_schema={
                "type": "object",
                "required": ["to", "body", "subject", "attachments"],
                "additionalProperties": False,
                "properties": {
                    "to": {
                        "type": "string",
                    },
                    "subject": {
                        "type": "string",
                    },
                    "body": {
                        "type": "string",
                    },
                    "attachments": {
                        "type": "array",
                        "description": "a list of paths from the room's storage of files to attach",
                        "items": {"type": "string"},
                    },
                },
            },
        )

    async def execute(
        self,
        context: ToolContext,
        *,
        to: str,
        subject: str,
        body: str,
        attachments: list[str],
    ):
        attachment_data = list[FileContent]()

        for attachment in attachments:
            try:
                attachment_data.append(
                    await context.room.storage.download(path=attachment)
                )
            except Exception as ex:
                logger.error(f"Unable to download file {ex}", exc_info=ex)
                raise RoomException(
                    f"Could not download a file from the room with the path {attachment}. Are you sure the path is correct file?"
                )

        await self.agent.start_thread(
            to_address=to, subject=subject, body=body, attachments=attachment_data
        )
        return {}


class NewEmailThread(FunctionTool):
    def __init__(self, *, agent: "MailBot"):
        self.agent = agent
        super().__init__(
            name="new_email_thread",
            title="New Email Thread",
            description="Starts a new email thread that is managed by the mailbot",
            input_schema={
                "type": "object",
                "required": ["to", "body", "subject"],
                "additionalProperties": False,
                "properties": {
                    "to": {
                        "type": "string",
                    },
                    "subject": {
                        "type": "string",
                    },
                    "body": {
                        "type": "string",
                    },
                },
            },
        )

    async def execute(
        self,
        context: ToolContext,
        *,
        to: str,
        subject: str,
        body: str,
    ):
        await self.agent.start_thread(
            to_address=to, subject=subject, body=body, attachments=[]
        )
        return {}


class MailBot(Worker):
    def __init__(
        self,
        *,
        queue: Optional[str] = None,
        name=None,
        title=None,
        description=None,
        requires=None,
        llm_adapter,
        toolkits=None,
        rules=None,
        email_address: str,
        domain: str = os.getenv("MESHAGENT_MAIL_DOMAIN", "mail.meshagent.com"),
        smtp: Optional[SmtpConfiguration] = None,
        toolkit_name: Optional[str] = None,
        whitelist: Optional[list[str]] = None,
        reply_all: bool = False,
        enable_attachments: bool = True,
        skill_dirs: Optional[list[str]] = None,
        annotations: Optional[list[str]] = None,
    ):
        if smtp is None:
            smtp = SmtpConfiguration()

        if queue is None:
            queue = email_address

        self._domain = domain
        self._smtp = smtp
        self._reply_all = reply_all
        self._enable_attachments = enable_attachments

        super().__init__(
            queue=queue,
            name=name,
            title=title,
            description=description,
            requires=requires,
            llm_adapter=llm_adapter,
            toolkits=toolkits,
            annotations=annotations,
            rules=rules
            or [
                "You MUST reply with plain text or markdown, do not reply in JSON format or HTML format"
            ],
        )
        self._email_address = email_address
        self._whitelist = whitelist

        if toolkit_name is not None:
            logger.info(f"mailbox will start toolkit {toolkit_name}")
            self._toolkit = RemoteToolkit(
                name=toolkit_name,
                tools=[
                    NewEmailThreadWithAttachments(agent=self)
                    if enable_attachments
                    else NewEmailThread(agent=self),
                ],
            )
        else:
            self._toolkit = None

        self._skill_dirs = skill_dirs

    def get_requirements(self):
        return [
            *super().get_requirements(),
            RequiredTable(
                name="emails",
                schema={"id": TextDataType(), "json": TextDataType()},
                scalar_indexes=["id"],
            ),
        ]

    async def load_message(self, *, message_id: str) -> dict | None:
        room = self.room
        messages = await room.database.search(table="emails", where={"id": message_id})

        if len(messages) == 0:
            return None

        return json.loads(messages[0]["json"])

    def message_to_json(self, *, message: EmailMessage, role: "MessageRole") -> dict:
        return mail_common.message_to_json(message=message, role=role)

    async def save_email_message(self, *, content: bytes, role: MessageRole) -> dict:
        room = self.room
        message = message_from_bytes(content, policy=default)

        now = datetime.now(timezone.utc)

        folder_path = (
            now.strftime("%Y/%m/%d")
            + "/"
            + now.strftime("%H/%M/%S")
            + "/"
            + secrets.token_hex(3)
        )

        queued_message = self.message_to_json(message=message, role=role)
        message_id = queued_message["id"]

        queued_message["role"] = role

        queued_message["path"] = f".emails/{folder_path}/message.json"

        for fname, bin_data in mail_common.iter_message_attachments(message):
            path = f".emails/{folder_path}/attachments/{fname}"
            logger.info(f"writing content to {path}")
            await room.storage.upload(path=path, data=bin_data)

            queued_message["attachments"].append(path)

        logger.info(f"received mail, {queued_message}")

        # write email
        path = f".emails/{folder_path}/message.eml"
        logger.info(f"writing source message.eml to {path}")
        await room.storage.upload(path=path, data=content)

        path = f".emails/{folder_path}/message.json"
        logger.info(f"writing source message.json to {path}")
        await room.storage.upload(
            path=path,
            data=json.dumps(queued_message, indent=4).encode("utf-8"),
        )

        await room.database.insert(
            table="emails",
            records=[{"id": message_id, "json": json.dumps(queued_message)}],
        )

        return queued_message

    async def load_thread(self, *, message: dict, thread: list[dict]):
        in_reply_to = message.get("in_reply_to", None)
        if in_reply_to is not None:
            source = await self.load_message(message_id=in_reply_to)

            if source is not None:
                thread.insert(0, source)

                await self.load_thread(message=source, thread=thread)

            else:
                logger.warning(f"message not found {in_reply_to}")

    async def append_message_context(
        self,
        *,
        message: dict,
        chat_context: AgentSessionContext,
        thread: list[dict],
    ):
        for msg in thread:
            if msg["role"] == "agent":
                chat_context.append_assistant_message(json.dumps(msg))

            else:
                chat_context.append_user_message(json.dumps(msg))

    async def get_rules(self):
        rules = [*self._rules]

        if self._skill_dirs is not None and len(self._skill_dirs) > 0:
            rules.append(
                "You have access to to following skills which follow the agentskills spec:"
            )
            rules.append(await to_prompt([*(Path(p) for p in self._skill_dirs)]))
            rules.append(
                "Use the shell or storage tool to find out more about skills and execute them when they are required"
            )

        return rules

    async def should_reply(self, *, message: dict) -> bool:
        should_reply, reason = mail_common.should_reply_to_message(
            message=message,
            email_address=self._email_address,
            whitelist=self._whitelist,
        )
        if not should_reply and reason == "message is not addressed to this mailbox":
            logger.warn(
                f"message not addressed to {self._email_address}, message will be ignored by the mailbot; "
                f"to={message.get('to')!r} cc={message.get('cc')!r} delivered_to={((message.get('meta') or {}).get('delivered_to') or '')!r}"
            )
            return False
        if not should_reply and reason is not None:
            logger.info(f"discarding message (error/auto-reply): {reason}")
            return False
        return should_reply

    async def process_message(
        self,
        *,
        chat_context: AgentSessionContext,
        message: dict,
        toolkits: list[Toolkit],
    ):
        logger.info("received a mail message")

        rules = await self.get_rules()

        logger.info(f"using rules {rules}")

        chat_context.replace_rules(rules)

        message_bytes = base64.b64decode(message["base64"])

        message = await self.save_email_message(content=message_bytes, role="user")

        if not await self.should_reply(message=message):
            return

        thread = [message]

        await self.load_thread(message=message, thread=thread)

        await self.append_message_context(
            message=message, chat_context=chat_context, thread=thread
        )

        thread_context = MailThreadContext(
            chat=chat_context, message=message, thread=thread
        )
        toolkits = await self.get_thread_toolkits(thread_context=thread_context)

        attachment_data = []

        try:
            if self._enable_attachments:

                class AttachTool(FunctionTool):
                    def __init__(self):
                        super().__init__(
                            name="attach file",
                            description="attach a file from the room to the conversation",
                            input_schema={
                                "type": "object",
                                "additionalProperties": False,
                                "required": ["path"],
                                "properties": {
                                    "path": {
                                        "type": "string",
                                        "description": "a path to a file in the room's storage",
                                    }
                                },
                            },
                        )

                    async def execute(self, context: ToolContext, *, path: str):
                        try:
                            storage_toolkits = [
                                t for t in toolkits if isinstance(t, StorageToolkit)
                            ]

                            if len(storage_toolkits) > 0:
                                attachment_data.append(
                                    await storage_toolkits[0].read_file(
                                        context=context, path=path
                                    )
                                )
                            else:
                                attachment_data.append(
                                    await context.room.storage.download(path=path)
                                )
                        except Exception as ex:
                            logger.error(f"Unable to download file {ex}", exc_info=ex)
                            raise RoomException(
                                f"Could not download a file from the room with the path {path}. Are you sure the path is correct file?"
                            )

                toolkits = [
                    *toolkits,
                    Toolkit(name="attachments", tools=[AttachTool()]),
                ]

            reply = await self._llm_adapter.next(
                context=chat_context,
                room=self.room,
                toolkits=toolkits,
            )
        except Exception as ex:
            logger.error(f"error while processing message {ex}", exc_info=ex)
            reply = "An error occurred while processing your message, please try again."

        logger.info(f"replying: {reply}")

        return await self.send_reply_message(
            message=message, reply=reply, attachments=attachment_data
        )

    def render_markdown(self, body: str):
        return mail_common.render_mail_markdown(body=body)

    def create_email_message(
        self,
        *,
        to_address: str,
        from_address: str,
        subject: str,
        body: str,
        correlation_id: Optional[str] = None,
    ) -> EmailMessage:
        return mail_common.create_email_message(
            to_address=to_address,
            from_address=from_address,
            subject=subject,
            body=body,
            correlation_id=correlation_id,
        )

    async def start(self, *, room: RoomClient):
        await super().start(room=room)
        if self._toolkit is not None:
            await self._toolkit.start(room=room)

    async def stop(self):
        if self._toolkit is not None:
            await self._toolkit.stop()
        await super().stop()

    async def start_thread(
        self,
        *,
        to_address: str,
        subject: str,
        body: str,
        from_address: Optional[str] = None,
        attachments: Optional[list[FileContent]] = None,
    ):
        msg = self.create_email_message(
            to_address=to_address,
            from_address=from_address or self._email_address,
            subject=subject,
            body=body,
        )

        reply_msg_dict = await self.save_email_message(
            content=msg.as_bytes(), role="agent"
        )

        if attachments is not None:
            reply_msg_dict["attachments"] = [*(x.name for x in attachments)]

            for attachment in attachments:
                maintype, subtype = attachment.mime_type.split("/")
                msg.add_attachment(
                    attachment.data,
                    maintype=maintype,
                    subtype=subtype,
                    filename=attachment.name,
                )

        logger.info(f"starting thread with message {reply_msg_dict}")

        username = self._smtp.username
        if username is None:
            username = self.room.local_participant.get_attribute("name")

        password = self._smtp.password
        if password is None:
            password = self.room.protocol.token

        hostname = self._smtp.hostname
        if hostname is None:
            hostname = self._domain

        port = self._smtp.port

        logger.info(f"using smtp {username}@{hostname}:{port}")

        await aiosmtplib.send(
            msg,
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            local_hostname=self._smtp.effective_local_hostname(),
        )

    def create_reply_email_message(
        self,
        *,
        message: dict,
        from_address: str,
        body: str,
        reply_all: bool = False,  # <-- choose behavior
    ) -> EmailMessage:
        return mail_common.create_reply_email_message(
            message=message,
            from_address=from_address,
            body=body,
            email_address=self._email_address,
            reply_all=reply_all,
        )

    async def send_reply_message(
        self,
        *,
        message: dict,
        reply: str,
        attachments: Optional[list[FileContent]] = None,
    ):
        msg = self.create_reply_email_message(
            message=message,
            from_address=self._email_address,
            body=reply,
            reply_all=self._reply_all,
        )

        reply_msg_dict = await self.save_email_message(
            content=msg.as_bytes(), role="agent"
        )

        if attachments is not None:
            reply_msg_dict["attachments"] = [*(x.name for x in attachments)]

            for attachment in attachments:
                maintype, subtype = attachment.mime_type.split("/")
                msg.add_attachment(
                    attachment.data,
                    maintype=maintype,
                    subtype=subtype,
                    filename=attachment.name,
                )

        logger.info(f"replying with message {reply_msg_dict}")

        username = self._smtp.username
        if username is None:
            username = self.room.local_participant.get_attribute("name")

        password = self._smtp.password
        if password is None:
            password = self.room.protocol.token

        hostname = self._smtp.hostname
        if hostname is None:
            hostname = self._domain

        port = self._smtp.port

        logger.info(f"using smtp {username}@{hostname}:{port}")

        await aiosmtplib.send(
            msg,
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            local_hostname=self._smtp.effective_local_hostname(),
        )

    async def get_thread_toolkits(
        self,
        *,
        thread_context: MailThreadContext,
    ) -> list[Toolkit]:
        toolkits = await self.get_required_toolkits(
            context=ToolContext(
                room=self.room,
                caller=self.room.local_participant,
                caller_context={"chat": thread_context.chat.to_json()},
            )
        )

        return [*self._toolkits, *toolkits]



import asyncio
from typing import Literal

import pytest
from pydantic import BaseModel

from meshagent.agents.messages import (
    AGENT_EVENT_TEXT_CONTENT_DELTA,
    AGENT_EVENT_TEXT_CONTENT_ENDED,
    AGENT_EVENT_TEXT_CONTENT_STARTED,
    AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
    AGENT_EVENT_TURN_ENDED,
    AGENT_EVENT_TURN_STARTED,
    AGENT_MESSAGE_TOOL_CALL_REJECT,
    AgentError,
    AgentTextContentDelta,
    AgentTextContentEnded,
    AgentTextContentStarted,
    AgentToolCallApprovalRequested,
    RejectAgentToolCall,
    TurnEnded,
    TurnStart,
    TurnStarted,
)
from meshagent.agents.process import Message
from meshagent.agents.toolkit_channel import ToolkitChannel
from meshagent.api import Participant
from meshagent.api.messaging import JsonContent, TextContent
from meshagent.api.room_server_client import RoomException
from meshagent.tools import ToolContext, ToolkitBuilder


class _FakeLocalParticipant(Participant):
    def __init__(self) -> None:
        super().__init__(id="assistant-id", attributes={"name": "assistant"})

    async def set_attribute(self, name: str, value) -> None:
        self._attributes[name] = value


class _FakeStorage:
    async def exists(self, *, path: str) -> bool:
        del path
        return False


class _FakeRoom:
    def __init__(self) -> None:
        self.local_participant = _FakeLocalParticipant()
        self.storage = _FakeStorage()


class _RecordingSupervisor:
    def __init__(self) -> None:
        self.sent: list[Message] = []

    def send(self, message: Message) -> None:
        self.sent.append(message)


class _FakeParticipant(Participant):
    def __init__(self, *, name: str, participant_id: str = "caller-id") -> None:
        super().__init__(id=participant_id, attributes={"name": name})


class _SearchToolkitConfig(BaseModel):
    name: Literal["search"] = "search"


class _FakeToolkitBuilder(ToolkitBuilder):
    def __init__(self) -> None:
        super().__init__(name="search", type=_SearchToolkitConfig)

    async def make(self, *, room, model: str, config: _SearchToolkitConfig):
        del room
        del model
        del config
        raise AssertionError("builder should not be called in this test")


async def _drain() -> None:
    await asyncio.sleep(0)
    await asyncio.sleep(0)


@pytest.mark.asyncio
async def test_toolkit_channel_exposes_remote_tool_and_returns_final_text() -> None:
    room = _FakeRoom()
    caller = _FakeParticipant(name="Caller")
    supervisor = _RecordingSupervisor()
    channel = ToolkitChannel(
        room=room,
        toolkit_name="assistant",
        thread_dir="/threads/tasks",
        toolkit_builders=[_FakeToolkitBuilder()],
    )
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        remote_toolkit = channel.make_remote_toolkit()
        assert remote_toolkit.name == "assistant"
        assert {tool.name for tool in remote_toolkit.tools} == {"run_assistant_task"}

        result_task = asyncio.create_task(
            remote_toolkit.invoke(
                context=ToolContext(room=room, caller=caller),
                name="run_assistant_task",
                input=JsonContent(
                    json={
                        "prompt": "Plan the work",
                        "tools": [{"name": "search"}],
                        "model": "gpt-5.4",
                        "instructions": "Be brief",
                    }
                ),
            )
        )
        await _drain()

        assert len(supervisor.sent) == 1
        request = supervisor.sent[0]
        assert isinstance(request.data, TurnStart)
        assert request.sender is caller
        assert request.data.thread_id == "/threads/tasks/Plan the work.thread"
        assert request.data.content[0].text == "Plan the work"
        assert request.data.toolkits == [{"name": "search"}]
        assert request.data.model == "gpt-5.4"
        assert request.data.instructions == "Be brief"

        channel.send(
            Message(
                data=TurnStarted(
                    type=AGENT_EVENT_TURN_STARTED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-1",
                    source_message_id=request.data.message_id,
                )
            )
        )
        channel.send(
            Message(
                data=AgentTextContentStarted(
                    type=AGENT_EVENT_TEXT_CONTENT_STARTED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-1",
                    item_id="text-1",
                )
            )
        )
        channel.send(
            Message(
                data=AgentTextContentDelta(
                    type=AGENT_EVENT_TEXT_CONTENT_DELTA,
                    thread_id=request.data.thread_id,
                    turn_id="turn-1",
                    item_id="text-1",
                    text="Hello",
                )
            )
        )
        channel.send(
            Message(
                data=AgentTextContentDelta(
                    type=AGENT_EVENT_TEXT_CONTENT_DELTA,
                    thread_id=request.data.thread_id,
                    turn_id="turn-1",
                    item_id="text-1",
                    text=" world",
                )
            )
        )
        channel.send(
            Message(
                data=AgentTextContentEnded(
                    type=AGENT_EVENT_TEXT_CONTENT_ENDED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-1",
                    item_id="text-1",
                )
            )
        )
        channel.send(
            Message(
                data=TurnEnded(
                    type=AGENT_EVENT_TURN_ENDED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-1",
                    error=None,
                )
            )
        )

        response = await result_task
        assert isinstance(response, TextContent)
        assert response.text == "Hello world"
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_toolkit_channel_respects_explicit_thread_id_and_surfaces_turn_errors() -> (
    None
):
    room = _FakeRoom()
    caller = _FakeParticipant(name="Caller")
    supervisor = _RecordingSupervisor()
    channel = ToolkitChannel(room=room, toolkit_name="assistant")
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        remote_toolkit = channel.make_remote_toolkit()
        result_task = asyncio.create_task(
            remote_toolkit.invoke(
                context=ToolContext(room=room, caller=caller),
                name="run_assistant_task",
                input=JsonContent(
                    json={
                        "prompt": "Continue",
                        "path": "/threads/shared.thread",
                        "thread_id": "/threads/shared.thread",
                    }
                ),
            )
        )
        await _drain()

        assert len(supervisor.sent) == 1
        request = supervisor.sent[0]
        assert isinstance(request.data, TurnStart)
        assert request.data.thread_id == "/threads/shared.thread"

        channel.send(
            Message(
                data=TurnStarted(
                    type=AGENT_EVENT_TURN_STARTED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-2",
                    source_message_id=request.data.message_id,
                )
            )
        )
        channel.send(
            Message(
                data=TurnEnded(
                    type=AGENT_EVENT_TURN_ENDED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-2",
                    error=AgentError(message="turn failed", code="RoomException"),
                )
            )
        )

        with pytest.raises(RoomException, match="turn failed"):
            await result_task
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_toolkit_channel_auto_rejects_tool_approvals() -> None:
    room = _FakeRoom()
    caller = _FakeParticipant(name="Caller")
    supervisor = _RecordingSupervisor()
    channel = ToolkitChannel(room=room, toolkit_name="assistant")
    await channel.start(supervisor)  # type: ignore[arg-type]
    try:
        remote_toolkit = channel.make_remote_toolkit()
        result_task = asyncio.create_task(
            remote_toolkit.invoke(
                context=ToolContext(room=room, caller=caller),
                name="run_assistant_task",
                input=JsonContent(json={"prompt": "Do the work"}),
            )
        )
        await _drain()

        request = supervisor.sent[0]
        assert isinstance(request.data, TurnStart)
        channel.send(
            Message(
                data=TurnStarted(
                    type=AGENT_EVENT_TURN_STARTED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-3",
                    source_message_id=request.data.message_id,
                )
            )
        )
        channel.send(
            Message(
                data=AgentToolCallApprovalRequested(
                    type=AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-3",
                    item_id="approval-1",
                    toolkit="shell",
                    tool="shell",
                    arguments={"command": "echo hi"},
                )
            )
        )
        await _drain()

        assert len(supervisor.sent) == 2
        rejection = supervisor.sent[1]
        assert isinstance(rejection.data, RejectAgentToolCall)
        assert rejection.data.type == AGENT_MESSAGE_TOOL_CALL_REJECT
        assert rejection.data.turn_id == "turn-3"
        assert rejection.data.item_id == "approval-1"

        channel.send(
            Message(
                data=TurnEnded(
                    type=AGENT_EVENT_TURN_ENDED,
                    thread_id=request.data.thread_id,
                    turn_id="turn-3",
                    error=AgentError(message="tool rejected", code="rejected"),
                )
            )
        )

        with pytest.raises(RoomException, match="tool rejected"):
            await result_task
    finally:
        await channel.stop(supervisor)  # type: ignore[arg-type]



import asyncio
import logging
import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from meshagent.api import RoomClient
from meshagent.api.room_server_client import (
    BinaryDataType,
    ListDataType,
    StructDataType,
    TextDataType,
    TimestampDataType,
)

logger = logging.getLogger("images_database")
_IMAGE_SAVE_MAX_RETRIES = 6
_IMAGE_SAVE_RETRY_BASE_DELAY_SECONDS = 0.2


@dataclass(frozen=True)
class SavedImage:
    id: str
    mime_type: str
    created_at: str
    created_by: str
    annotations: dict[str, str]


class ImagesDatabase:
    TABLE_NAME = "images"
    _METADATA_COLUMNS = ["id", "mime_type", "created_at", "created_by", "annotations"]

    def __init__(self, room: RoomClient):
        self._room = room
        self._ready = False
        self._lock = asyncio.Lock()

    def _schema(self) -> dict:
        return {
            "id": TextDataType(nullable=False),
            "data": BinaryDataType(nullable=False),
            "mime_type": TextDataType(nullable=False),
            "created_at": TimestampDataType(nullable=False),
            "created_by": TextDataType(nullable=False),
            "annotations": ListDataType(
                element_type=StructDataType(
                    fields={
                        "key": TextDataType(nullable=False),
                        "value": TextDataType(),
                    }
                )
            ),
        }

    @staticmethod
    def _normalize_annotations(
        annotations: Optional[dict[str, str]],
    ) -> dict[str, str]:
        if annotations is None:
            return {}

        normalized: dict[str, str] = {}
        for key, value in annotations.items():
            normalized[str(key)] = str(value)
        return normalized

    @staticmethod
    def _encode_annotations(annotations: dict[str, str]) -> list[dict[str, str]]:
        return [{"key": key, "value": value} for key, value in annotations.items()]

    @staticmethod
    def _decode_annotations(value: Any) -> dict[str, str]:
        if value is None:
            return {}

        if isinstance(value, dict):
            return {str(k): str(v) for k, v in value.items()}

        if not isinstance(value, list):
            return {}

        decoded: dict[str, str] = {}
        for item in value:
            if not isinstance(item, dict):
                continue
            key = item.get("key")
            if not isinstance(key, str):
                continue
            item_value = item.get("value")
            decoded[key] = "" if item_value is None else str(item_value)
        return decoded

    @staticmethod
    def _to_saved_image(record: dict[str, Any]) -> SavedImage:
        return SavedImage(
            id=str(record.get("id") or ""),
            mime_type=str(record.get("mime_type") or "application/octet-stream"),
            created_at=str(record.get("created_at") or ""),
            created_by=str(record.get("created_by") or ""),
            annotations=ImagesDatabase._decode_annotations(record.get("annotations")),
        )

    @staticmethod
    def _is_commit_conflict_error(exc: Exception) -> bool:
        message = str(exc).lower()
        return (
            "commit conflict" in message
            and "concurrent transaction" in message
            and "lance" in message
        )

    async def _ensure_ready(self) -> None:
        if self._ready:
            return

        async with self._lock:
            if self._ready:
                return

            schema = self._schema()
            await self._room.database.create_table_with_schema(
                name=self.TABLE_NAME,
                schema=schema,
                mode="create_if_not_exists",
            )

            existing_schema = await self._room.database.inspect(table=self.TABLE_NAME)
            missing_columns = {
                key: value
                for key, value in schema.items()
                if key not in existing_schema
            }
            if len(missing_columns) > 0:
                await self._room.database.add_columns(
                    table=self.TABLE_NAME,
                    new_columns=missing_columns,
                )

            try:
                await self._room.database.create_scalar_index(
                    table=self.TABLE_NAME,
                    column="id",
                )
            except Exception:
                logger.debug(
                    "unable to create images.id scalar index; continuing",
                    exc_info=True,
                )

            self._ready = True

    async def save(
        self,
        *,
        data: bytes,
        mime_type: str,
        created_by: str,
        annotations: Optional[dict[str, str]] = None,
        image_id: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> SavedImage:
        await self._ensure_ready()

        normalized_annotations = self._normalize_annotations(annotations=annotations)
        normalized_id = image_id or str(uuid.uuid4())
        normalized_created_at = created_at or datetime.now(
            timezone.utc
        ).isoformat().replace("+00:00", "Z")

        record = {
            "id": normalized_id,
            "data": data,
            "mime_type": mime_type,
            "created_at": normalized_created_at,
            "created_by": created_by,
            "annotations": self._encode_annotations(normalized_annotations),
        }

        for attempt in range(_IMAGE_SAVE_MAX_RETRIES):
            try:
                await self._room.database.insert(
                    table=self.TABLE_NAME,
                    records=[record],
                )
                break
            except Exception as ex:
                if not self._is_commit_conflict_error(ex):
                    raise
                if attempt >= _IMAGE_SAVE_MAX_RETRIES - 1:
                    raise

                retry_delay = _IMAGE_SAVE_RETRY_BASE_DELAY_SECONDS * (2**attempt)
                logger.warning(
                    "images insert hit commit conflict; retrying (%d/%d) in %.1fs",
                    attempt + 1,
                    _IMAGE_SAVE_MAX_RETRIES,
                    retry_delay,
                )
                await asyncio.sleep(retry_delay)

        return SavedImage(
            id=normalized_id,
            mime_type=mime_type,
            created_at=normalized_created_at,
            created_by=created_by,
            annotations=normalized_annotations,
        )

    async def read(self, *, image_id: str) -> Optional[SavedImage]:
        await self._ensure_ready()

        rows = await self._room.database.search(
            table=self.TABLE_NAME,
            where={"id": image_id},
            limit=1,
            select=self._METADATA_COLUMNS,
        )
        if len(rows) == 0:
            return None

        return self._to_saved_image(rows[0])

    async def search(
        self,
        *,
        where: Optional[str | dict[str, Any]] = None,
        limit: Optional[int] = None,
    ) -> list[SavedImage]:
        await self._ensure_ready()

        rows = await self._room.database.search(
            table=self.TABLE_NAME,
            where=where,
            limit=limit,
            select=self._METADATA_COLUMNS,
        )
        return [self._to_saved_image(row) for row in rows]

    async def read_data(self, *, image_id: str) -> Optional[bytes]:
        await self._ensure_ready()

        rows = await self._room.database.search(
            table=self.TABLE_NAME,
            where={"id": image_id},
            limit=1,
            select=["data"],
        )
        if len(rows) == 0:
            return None

        data = rows[0].get("data")
        if isinstance(data, bytearray):
            return bytes(data)
        if isinstance(data, bytes):
            return data
        return None

    async def optimize(self) -> None:
        await self._ensure_ready()
        await self._room.database.optimize(table=self.TABLE_NAME)



from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from meshagent.agents.agent import SingleRoomAgent, AgentSessionContext
from meshagent.api.chan import Chan
from meshagent.api import (
    RoomMessage,
    RoomClient,
    RemoteParticipant,
    Participant,
    Requirement,
    Element,
    MeshDocument,
)
from meshagent.api.messaging import JsonContent
from meshagent.tools import (
    Toolkit,
    ToolContext,
    make_toolkits,
    ToolkitBuilder,
    ToolkitConfig,
    RemoteToolkit,
    FunctionTool,
)
from meshagent.agents.adapter import LLMAdapter
from meshagent.openai.tools.responses_adapter import (
    ReasoningTool,
    OpenAIResponsesAdapter,
    OpenAIResponsesSessionContext,
)
from meshagent.openai.tools.completions_adapter import OpenAICompletionsAdapter
from meshagent.agents.thread_adapter import ThreadAdapter
from meshagent.agents.completions_thread_adapter import CompletionsThreadAdapter
from meshagent.agents.responses_thread_adapter import (
    ResponsesThreadAdapter,
    response_event_to_agent_event,
)

import uuid
import posixpath
import re
from copy import deepcopy
from datetime import datetime, timedelta, timezone
import asyncio
from typing import Any, Optional, Callable, AsyncIterator, Awaitable, Literal
import logging
import warnings
from asyncio import CancelledError
from meshagent.api import RoomException
from meshagent.api.chan import ChanClosed
from opentelemetry import trace
import json
import aiohttp
from pydantic import BaseModel
from meshagent.tools import tool
from meshagent.tools.strict_schema import ensure_strict_json_schema
from pathlib import Path
from meshagent.agents.skills import to_prompt
from meshagent.agents.toolkit_schema import build_tools_property_schema
from meshagent.agents.thread_schema import thread_list_schema, thread_schema


tracer = trace.get_tracer("meshagent.chatbot")

logger = logging.getLogger("chat")
_legacy_chatbot_init_chat_context_warned: set[type] = set()
_DEFAULT_NEW_THREAD_NAME_RULES = [
    "generate a concise, friendly title for this chat thread",
    "return only a thread_name value suitable for display in a thread list",
    "thread_name should be 2-6 words and topic-focused",
    "use normal capitalization and spaces, and do not include a .thread extension",
]

ThreadStatusMode = Literal["busy", "steerable"]


class ChatAttachment(BaseModel):
    path: str


class ChatMessageInput(BaseModel):
    text: str
    attachments: Optional[list[ChatAttachment]] = None
    tools: Optional[list[Any]] = None


class ChatBotReasoningTool(ReasoningTool):
    def __init__(self, *, room: RoomClient, thread_context: "ChatThreadContext"):
        super().__init__()
        self.thread_context = thread_context
        self.room = room

        self._reasoning_element = None
        self._reasoning_item = None

    def _get_messages_element(self):
        messages = self.thread_context.thread.root.get_children_by_tag_name("messages")
        if len(messages) > 0:
            return messages[0]
        return None

    async def on_reasoning_summary_part_added(
        self,
        context: ToolContext,
        *,
        item_id: str,
        output_index: int,
        part: dict,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        el = self._get_messages_element()
        if el is None:
            logger.warning("missing messages element, cannot log reasoning")
        else:
            self._reasoning_element = el.append_child("reasoning", {"summary": ""})

    async def on_reasoning_summary_part_done(
        self,
        context: ToolContext,
        *,
        item_id: str,
        output_index: int,
        part: dict,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        self._reasoning_element = None

    async def on_reasoning_summary_text_delta(
        self,
        context: ToolContext,
        *,
        delta: str,
        output_index: int,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        el = self._reasoning_element
        el.set_attribute("summary", el.get_attribute("summary") + delta)

    async def on_reasoning_summary_text_done(
        self,
        context: ToolContext,
        *,
        item_id: str,
        output_index: int,
        sequence_number: int,
        summary_index: int,
        type: str,
        **extra,
    ):
        pass


def get_online_participants(
    *,
    room: RoomClient,
    thread: MeshDocument,
    exclude: Optional[list[Participant]] = None,
) -> list[RemoteParticipant]:
    results = list[RemoteParticipant]()

    for prop in thread.root.get_children():
        if prop.tag_name == "members":
            for member in prop.get_children():
                for online in room.messaging.get_participants():
                    if online.get_attribute("name") == member.get_attribute("name"):
                        if exclude is None or online not in exclude:
                            results.append(online)

    return results


class ChatThreadContext:
    def __init__(
        self,
        *,
        session: AgentSessionContext,
        thread: MeshDocument,
        path: str,
        participants: Optional[list[RemoteParticipant]] = None,
        event_handler: Optional[Callable[[dict], None]] = None,
    ):
        self.thread = thread
        if participants is None:
            participants = []

        self.participants = participants
        self.session = session
        self.path = path
        self._event_handler = event_handler

    async def __aenter__(self) -> "ChatThreadContext":
        await self.session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.session.__aexit__(exc_type, exc, tb)

    def emit(self, event: dict):
        if self._event_handler is not None:
            self._event_handler(event)

    @property
    def context_id(self) -> str:
        return self.session.id

    def to_caller_context(self) -> dict:
        return {"chat": self.session.to_json()}


class ChatBotClient:
    def __init__(
        self,
        *,
        room: RoomClient,
        participant_name: str,
        thread_path: str,
        timeout: float = 30,
    ):
        self.room = room
        self.participant_name = participant_name
        self.thread_path = thread_path
        self._messages: asyncio.Queue[str] = asyncio.Queue()
        self._doc = None
        self._participant: Optional[RemoteParticipant] = None
        self._timeout = timeout

    async def start(self) -> None:
        self._doc = await self.room.sync.open(path=self.thread_path)
        self.room.messaging.on("message", self._on_message)
        await asyncio.sleep(1)
        await self._wait_for_participant()
        if self._participant is not None:
            await self.room.messaging.send_message(
                to=self._participant,
                type="opened",
                message={"path": self.thread_path},
                attachment=None,
            )

    async def stop(self) -> None:
        await self.room.sync.close(path=self.thread_path)
        self.room.messaging.off("message", self._on_message)

    async def __aenter__(self) -> "ChatBotClient":
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, exc_tb) -> None:
        await self.stop()

    def _on_message(self, message: RoomMessage) -> None:
        received_type = message.type
        received_message = message.message

        if received_type == "chat" and received_message["path"] == self.thread_path:
            self._messages.put_nowait(received_message["text"])

    async def _wait_for_participant(self) -> None:
        try:
            async with asyncio.timeout(self._timeout):
                while self._participant is None:
                    for participant in self.room.messaging.get_participants():
                        if participant.get_attribute("name") == self.participant_name:
                            self._participant = participant
                            break

                    await asyncio.sleep(1)
        except asyncio.TimeoutError as exc:
            raise RoomException(
                f"timed out waiting for {self.participant_name}"
            ) from exc

    async def clear(self) -> None:
        if self._participant is None:
            return

        await self.room.messaging.send_message(
            to=self._participant,
            type="clear",
            message={"path": self.thread_path},
            attachment=None,
        )

    async def cancel(self) -> None:
        if self._participant is None:
            return

        await self.room.messaging.send_message(
            to=self._participant,
            type="cancel",
            message={"path": self.thread_path},
            attachment=None,
        )

    async def send_approval_decision(self, *, approval_id: str, approve: bool) -> None:
        if self._participant is None:
            return

        normalized_id = approval_id.strip()
        if normalized_id == "":
            raise RoomException("approval_id is required")

        await self.room.messaging.send_message(
            to=self._participant,
            type="approved" if approve else "rejected",
            message={
                "path": self.thread_path,
                "approval_id": normalized_id,
            },
            attachment=None,
        )

    async def send(
        self,
        *,
        text: str,
        tools: Optional[list[ToolkitConfig]] = None,
        attachments: Optional[list[dict]] = None,
        steer: bool = False,
    ) -> None:
        if self._participant is None or self._doc is None:
            raise RoomException("chat client not started")

        tool_payload = [tool.model_dump(mode="json") for tool in tools or []]
        attachment_payload: list[dict[str, str]] = []
        if attachments is not None:
            for attachment in attachments:
                if not isinstance(attachment, dict):
                    continue
                path = attachment.get("path")
                if not isinstance(path, str):
                    continue
                normalized_path = path.strip()
                if normalized_path == "":
                    continue
                attachment_payload.append({"path": normalized_path})

        await self.room.messaging.send_message(
            to=self._participant,
            type="steer" if steer else "chat",
            message={
                "text": text,
                "path": self.thread_path,
                "tools": tool_payload,
                "attachments": attachment_payload,
                "store": True,
            },
            attachment=None,
        )

    def __aiter__(self) -> AsyncIterator[str]:
        return self

    async def __anext__(self) -> str:
        return await self._messages.get()

    async def receive(self) -> str:
        return await self._messages.get()

    async def receive_nowait(self) -> str:
        return await self._messages.get_nowait()


@dataclass
class _QueuedChatMessage:
    type: str
    message: dict
    from_participant: RemoteParticipant
    result: asyncio.Future = field(default_factory=asyncio.Future)


class ChatBotBase(SingleRoomAgent, ABC):
    def __init__(
        self,
        *,
        name=None,
        title=None,
        description=None,
        requires: Optional[list[Requirement]] = None,
        toolkits: Optional[list[Toolkit]] = None,
        rules: Optional[list[str]] = None,
        client_rules: Optional[dict[str, list[str]]] = None,
        auto_greet_message: Optional[str] = None,
        empty_state_title: Optional[str] = None,
        annotations: Optional[list[str]] = None,
        always_reply: Optional[bool] = None,
        skill_dirs: Optional[list[str]] = None,
        thread_dir: Optional[str] = None,
        threading_mode: Optional[str] = None,
        thread_name_rules: Optional[list[str]] = None,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            annotations=annotations,
        )

        if toolkits is None:
            toolkits = []

        if always_reply is None:
            always_reply = False

        self._always_reply = always_reply

        self._message_channels = dict[str, Chan[_QueuedChatMessage]]()

        self._room: RoomClient | None = None
        self._toolkits = toolkits
        self._client_rules = client_rules

        if rules is None:
            rules = []

        self._rules = rules
        self._is_typing = dict[str, asyncio.Task]()
        self._auto_greet_message = auto_greet_message

        if empty_state_title is None:
            empty_state_title = "How can I help you?"
        self._empty_state_title = empty_state_title

        self._thread_tasks = dict[str, asyncio.Task]()
        self._open_threads = {}
        self._thread_contexts: dict[str, ChatThreadContext] = {}
        self._thread_list_document: Optional[MeshDocument] = None
        self._thread_list_path: Optional[str] = None

        self._skill_dirs = skill_dirs
        self._thread_status_values: dict[str, str] = {}
        self._thread_status_mode_values: dict[str, ThreadStatusMode] = {}
        self._thread_status_started_at_values: dict[str, str] = {}
        self._thread_status_keys: dict[str, str] = {}
        self._thread_status_locks: dict[str, asyncio.Lock] = {}
        self._thread_status_generations: dict[str, int] = {}
        self._threading_mode = (
            threading_mode.strip()
            if isinstance(threading_mode, str) and threading_mode.strip() != ""
            else None
        )
        self._thread_dir = (
            self._normalize_thread_dir(thread_dir=thread_dir)
            if thread_dir is not None
            else None
        )
        if thread_name_rules is not None and len(thread_name_rules) > 0:
            self._thread_name_rules = [*thread_name_rules]
        else:
            self._thread_name_rules = [*_DEFAULT_NEW_THREAD_NAME_RULES]

    def thread_name_adapter(self) -> Optional[LLMAdapter]:
        return None

    def _normalize_thread_dir(self, *, thread_dir: str) -> str:
        normalized = thread_dir.strip().rstrip("/")
        if normalized == "":
            raise RoomException("thread_dir must not be empty")
        return normalized

    def _default_thread_dir(self) -> str:
        local_name = "chat"
        if self._room is not None and self._room.local_participant is not None:
            participant_name = self._room.local_participant.get_attribute("name")
            if isinstance(participant_name, str) and participant_name.strip() != "":
                local_name = participant_name.strip()

        return self._normalize_thread_dir(
            thread_dir=posixpath.join(".threads", local_name)
        )

    def _get_thread_dir(self) -> str:
        if self._thread_dir is not None:
            return self._thread_dir
        return self._default_thread_dir()

    def _thread_list_dir(self) -> Optional[str]:
        if self._thread_dir is not None:
            return self._thread_dir
        if self._threading_mode == "default-new":
            return self._default_thread_dir()
        return None

    def _sanitize_thread_name(self, *, value: str) -> str:
        normalized = value.strip()
        if normalized.endswith(".thread"):
            normalized = normalized[: -len(".thread")]

        normalized = re.sub(r"[-_/]+", " ", normalized)
        normalized = re.sub(r"\s+", " ", normalized).strip(" .-_")
        normalized = re.sub(r"[^A-Za-z0-9 .,!?':()&]+", "", normalized)
        normalized = re.sub(r"\s+", " ", normalized).strip(" .-_")
        if normalized == "":
            return "New Chat"
        if normalized == normalized.lower() or normalized == normalized.upper():
            normalized = normalized.title()
        return normalized[:64].strip() or "New Chat"

    def _fallback_thread_name(self, *, text: str) -> str:
        normalized_text = text.strip()
        if normalized_text != "":
            return self._sanitize_thread_name(value=normalized_text)
        return "New Chat"

    def _thread_path_for_name(self, *, thread_name: str, thread_dir: str) -> str:
        return posixpath.join(thread_dir, f"{thread_name}.thread")

    def _utc_now_iso(self) -> str:
        return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")

    def _thread_list_index_path(self) -> Optional[str]:
        thread_dir = self._thread_list_dir()
        if thread_dir is None:
            return None
        return posixpath.join(thread_dir, "index.threadl")

    def _thread_list_entry_name_for_path(self, *, path: str) -> str:
        filename = posixpath.basename(path.strip())
        if filename.endswith(".thread"):
            filename = filename[: -len(".thread")]
        raw_name = filename.strip()
        if raw_name != "":
            try:
                parsed_uuid = uuid.UUID(raw_name)
                if str(parsed_uuid) == raw_name.lower():
                    return "New Chat"
            except ValueError:
                pass

        normalized = self._sanitize_thread_name(value=filename)
        return normalized

    def _find_thread_list_entry(self, *, path: str) -> Optional[Element]:
        if self._thread_list_dir() is None:
            return None

        if self._thread_list_document is None:
            return None

        for child in self._thread_list_document.root.get_children():
            if child.tag_name != "thread":
                continue
            if child.get_attribute("path") == path:
                return child

        return None

    def _upsert_thread_list_entry(
        self,
        *,
        path: str,
        name: Optional[str] = None,
        created_at: Optional[str] = None,
        modified_at: Optional[str] = None,
    ) -> None:
        if self._thread_list_dir() is None or self._thread_list_document is None:
            return

        now = self._utc_now_iso()
        provided_name = name.strip() if isinstance(name, str) else ""

        entry = self._find_thread_list_entry(path=path)
        if entry is None:
            resolved_name = (
                provided_name
                if provided_name != ""
                else self._thread_list_entry_name_for_path(path=path)
            )
            created_value = (
                created_at.strip()
                if isinstance(created_at, str) and created_at.strip() != ""
                else now
            )
            modified_value = (
                modified_at.strip()
                if isinstance(modified_at, str) and modified_at.strip() != ""
                else created_value
            )
            self._thread_list_document.root.append_child(
                tag_name="thread",
                attributes={
                    "name": resolved_name,
                    "path": path,
                    "created_at": created_value,
                    "modified_at": modified_value,
                },
            )
            return

        if provided_name != "":
            entry.set_attribute("name", provided_name)
        else:
            existing_name = entry.get_attribute("name")
            if not isinstance(existing_name, str) or existing_name.strip() == "":
                entry.set_attribute(
                    "name",
                    self._thread_list_entry_name_for_path(path=path),
                )

        entry.set_attribute("path", path)

        existing_created_at = entry.get_attribute("created_at")
        created_value = (
            existing_created_at.strip()
            if isinstance(existing_created_at, str)
            and existing_created_at.strip() != ""
            else (
                created_at.strip()
                if isinstance(created_at, str) and created_at.strip() != ""
                else now
            )
        )
        entry.set_attribute("created_at", created_value)

        modified_value = self._next_thread_list_modified_at(
            entry=entry,
            modified_at=modified_at,
        )
        entry.set_attribute("modified_at", modified_value)

    def _record_new_thread_in_index(
        self, *, path: str, name: Optional[str] = None
    ) -> None:
        now = self._utc_now_iso()
        self._upsert_thread_list_entry(
            path=path,
            name=name,
            created_at=now,
            modified_at=now,
        )

    def _touch_thread_in_index(self, *, path: str) -> None:
        self._upsert_thread_list_entry(
            path=path,
            modified_at=self._utc_now_iso(),
        )

    def _next_thread_list_modified_at(
        self,
        *,
        entry: Element | None,
        modified_at: str | None = None,
    ) -> str:
        candidate = (
            modified_at.strip()
            if isinstance(modified_at, str) and modified_at.strip() != ""
            else self._utc_now_iso()
        )
        if entry is None:
            return candidate

        existing_modified_at = self._parse_iso_datetime(
            value=entry.get_attribute("modified_at")
        )
        candidate_modified_at = self._parse_iso_datetime(value=candidate)
        if (
            existing_modified_at is None
            or candidate_modified_at is None
            or candidate_modified_at > existing_modified_at
        ):
            return candidate

        return (
            (existing_modified_at + timedelta(microseconds=1))
            .isoformat()
            .replace(
                "+00:00",
                "Z",
            )
        )

    def _thread_list_entries(self) -> list[Element]:
        if self._thread_list_dir() is None or self._thread_list_document is None:
            return []

        return [
            child
            for child in self._thread_list_document.root.get_children()
            if child.tag_name == "thread"
        ]

    def _parse_iso_datetime(self, *, value: Any) -> Optional[datetime]:
        if not isinstance(value, str):
            return None

        raw = value.strip()
        if raw == "":
            return None

        normalized = raw[:-1] + "+00:00" if raw.endswith("Z") else raw
        try:
            parsed = datetime.fromisoformat(normalized)
        except ValueError:
            return None

        if parsed.tzinfo is None:
            return parsed.replace(tzinfo=timezone.utc)
        return parsed

    def _thread_sort_datetime(self, *, entry: Element) -> datetime:
        modified_at = self._parse_iso_datetime(value=entry.get_attribute("modified_at"))
        if modified_at is not None:
            return modified_at
        created_at = self._parse_iso_datetime(value=entry.get_attribute("created_at"))
        if created_at is not None:
            return created_at
        return datetime.min.replace(tzinfo=timezone.utc)

    def _sorted_thread_list_entries(self) -> list[Element]:
        return sorted(
            self._thread_list_entries(),
            key=lambda entry: self._thread_sort_datetime(entry=entry),
            reverse=True,
        )

    def _thread_list_slice(
        self,
        *,
        entries: list[Element],
        limit: int,
        offset: int,
    ) -> list[Element]:
        normalized_offset = max(0, int(offset))
        normalized_limit = max(1, min(200, int(limit)))
        return entries[normalized_offset : normalized_offset + normalized_limit]

    async def _open_thread_list_document(self) -> None:
        index_path = self._thread_list_index_path()
        if index_path is None or self._room is None:
            return

        if (
            self._thread_list_document is not None
            and self._thread_list_path == index_path
        ):
            return

        self._thread_list_path = index_path
        try:
            self._thread_list_document = await self.room.sync.open(
                path=index_path,
                schema=thread_list_schema,
            )
        except Exception as ex:
            logger.warning(
                "unable to open thread list document at %s",
                index_path,
                exc_info=ex,
            )
            self._thread_list_document = None
            self._thread_list_path = None

    async def _close_thread_list_document(
        self,
        *,
        room: Optional[RoomClient] = None,
    ) -> None:
        if self._thread_list_document is None or self._thread_list_path is None:
            return

        close_room = room if room is not None else self._room
        close_path = self._thread_list_path
        self._thread_list_document = None
        self._thread_list_path = None

        if close_room is None:
            return

        try:
            await close_room.sync.close(path=close_path)
        except Exception as ex:
            logger.warning(
                "unable to close thread list document at %s",
                close_path,
                exc_info=ex,
            )

    async def _next_available_thread_path(self, *, base_path: str) -> str:
        if self._room is None:
            return base_path

        try:
            exists = await self._room.storage.exists(path=base_path)
        except Exception:
            return base_path

        if not exists:
            return base_path

        thread_dir, filename = posixpath.split(base_path)
        if filename.endswith(".thread"):
            base_name = filename[: -len(".thread")]
        else:
            base_name = filename

        for index in range(2, 1000):
            candidate = posixpath.join(thread_dir, f"{base_name} {index}.thread")
            try:
                if not await self._room.storage.exists(path=candidate):
                    return candidate
            except Exception:
                return candidate

        return posixpath.join(thread_dir, f"{base_name}-{uuid.uuid4().hex[:8]}.thread")

    async def _generate_thread_friendly_name(
        self,
        *,
        context: ToolContext,
        text: str,
    ) -> str:
        generated_name = self._fallback_thread_name(text=text)
        adapter = self.thread_name_adapter()
        if adapter is not None:
            caller_context_json = context.caller_context
            chat_context_json = None
            if isinstance(caller_context_json, dict):
                candidate = caller_context_json.get("chat")
                if isinstance(candidate, dict):
                    chat_context_json = candidate

            session = adapter.create_session()
            if chat_context_json is not None:
                caller_context = AgentSessionContext.from_json(chat_context_json)
                session.messages.extend(deepcopy(caller_context.messages))
                session.previous_messages.extend(
                    deepcopy(caller_context.previous_messages)
                )
                session.previous_response_id = caller_context.previous_response_id
            cloned_context = session.copy()
            async with cloned_context:
                cloned_context.replace_rules(rules=self._thread_name_rules)
                cloned_context.append_user_message(text)
                try:
                    response = await adapter.next(
                        context=cloned_context,
                        room=context.room,
                        model=self.default_model(),
                        on_behalf_of=context.on_behalf_of or context.caller,
                        toolkits=[],
                        output_schema={
                            "type": "object",
                            "required": ["thread_name"],
                            "additionalProperties": False,
                            "properties": {
                                "thread_name": {
                                    "type": "string",
                                    "description": "2-6 word topic name for the task thread",
                                }
                            },
                        },
                    )
                    if isinstance(response, dict):
                        thread_name = response.get("thread_name")
                        if isinstance(thread_name, str):
                            generated_name = self._sanitize_thread_name(
                                value=thread_name
                            )
                except Exception as ex:
                    logger.warning(
                        "unable to auto-generate chat thread name, using fallback",
                        exc_info=ex,
                    )
        return generated_name

    async def _generate_new_thread_info(
        self,
        *,
        context: ToolContext,
        text: str,
    ) -> tuple[str, str]:
        friendly_name = await self._generate_thread_friendly_name(
            context=context,
            text=text,
        )
        guid_name = str(uuid.uuid4())
        path = self._thread_path_for_name(
            thread_name=guid_name,
            thread_dir=self._get_thread_dir(),
        )
        path = await self._next_available_thread_path(base_path=path)
        return path, friendly_name

    def _thread_status_attribute_name(self, *, path: str) -> str:
        return f"thread.status.{path}"

    def _thread_status_text_attribute_name(self, *, path: str) -> str:
        return f"thread.status.text.{path}"

    def _thread_status_mode_attribute_name(self, *, path: str) -> str:
        return f"thread.status.mode.{path}"

    def _thread_status_started_at_attribute_name(self, *, path: str) -> str:
        return f"thread.status.started_at.{path}"

    def _status_lock(self, *, path: str) -> asyncio.Lock:
        lock = self._thread_status_locks.get(path)
        if lock is None:
            lock = asyncio.Lock()
            self._thread_status_locks[path] = lock
        return lock

    def _next_thread_status_generation(self, *, path: str) -> int:
        generation = self._thread_status_generations.get(path, 0) + 1
        self._thread_status_generations[path] = generation
        return generation

    def _normalize_thread_status_mode(
        self, *, mode: Optional[str]
    ) -> Optional[ThreadStatusMode]:
        if mode is None:
            return None
        normalized = mode.strip().lower()
        if normalized == "":
            return None
        if normalized not in ("busy", "steerable"):
            raise RoomException(f"unsupported thread status mode '{mode}'")
        if normalized == "busy":
            return "busy"
        return "steerable"

    def processing_thread_status_mode(
        self, *, path: str, thread_context: Optional["ChatThreadContext"]
    ) -> ThreadStatusMode:
        del path
        del thread_context
        return "busy"

    async def set_thread_status(
        self,
        *,
        path: str,
        status: Optional[str],
        mode: Optional[str] = None,
    ) -> None:
        if self._room is None or self._room.local_participant is None:
            return

        async def set_local_attribute(
            attribute_name: str, value: Optional[str]
        ) -> None:
            try:
                await self._room.local_participant.set_attribute(attribute_name, value)
            except ChanClosed:
                logger.debug(
                    "room channel closed while setting thread status '%s'",
                    attribute_name,
                )

        attribute_name = self._thread_status_attribute_name(path=path)
        text_attribute_name = self._thread_status_text_attribute_name(path=path)
        mode_attribute_name = self._thread_status_mode_attribute_name(path=path)
        started_at_attribute_name = self._thread_status_started_at_attribute_name(
            path=path
        )
        if status is None:
            self._thread_status_values.pop(path, None)
            self._thread_status_mode_values.pop(path, None)
            self._thread_status_started_at_values.pop(path, None)
            await set_local_attribute(attribute_name, None)
            await set_local_attribute(text_attribute_name, None)
            await set_local_attribute(mode_attribute_name, None)
            await set_local_attribute(started_at_attribute_name, None)
            return

        normalized = status.strip()
        if normalized == "":
            self._thread_status_values.pop(path, None)
            self._thread_status_mode_values.pop(path, None)
            self._thread_status_started_at_values.pop(path, None)
            await set_local_attribute(attribute_name, None)
            await set_local_attribute(text_attribute_name, None)
            await set_local_attribute(mode_attribute_name, None)
            await set_local_attribute(started_at_attribute_name, None)
            return

        normalized_mode = self._normalize_thread_status_mode(mode=mode)
        if normalized_mode is None:
            normalized_mode = self._thread_status_mode_values.get(path, "busy")

        started_at = self._thread_status_started_at_values.get(path)
        if started_at is None:
            started_at = datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")

        if (
            self._thread_status_values.get(path) == normalized
            and self._thread_status_mode_values.get(path) == normalized_mode
            and self._thread_status_started_at_values.get(path) == started_at
        ):
            return

        self._thread_status_values[path] = normalized
        self._thread_status_mode_values[path] = normalized_mode
        self._thread_status_started_at_values[path] = started_at
        await set_local_attribute(attribute_name, normalized)
        await set_local_attribute(text_attribute_name, normalized)
        await set_local_attribute(mode_attribute_name, normalized_mode)
        await set_local_attribute(started_at_attribute_name, started_at)

    async def _apply_thread_status(
        self,
        *,
        path: str,
        status: Optional[str],
        generation: Optional[int] = None,
    ) -> None:
        lock = self._status_lock(path=path)
        async with lock:
            if generation is not None:
                current_generation = self._thread_status_generations.get(path, 0)
                if generation != current_generation:
                    return
            await self.set_thread_status(path=path, status=status)

    def _set_thread_status_nowait(self, *, path: str, status: Optional[str]) -> None:
        generation = self._next_thread_status_generation(path=path)

        async def run() -> None:
            try:
                await self._apply_thread_status(
                    path=path,
                    status=status,
                    generation=generation,
                )
            except Exception as ex:
                logger.error(
                    f"unable to set thread status for {path}",
                    exc_info=ex,
                )

        asyncio.create_task(run())

    async def clear_thread_status(self, *, path: str) -> None:
        self._thread_status_keys.pop(path, None)
        generation = self._next_thread_status_generation(path=path)
        await self._apply_thread_status(
            path=path,
            status=None,
            generation=generation,
        )

    def _clear_thread_status_nowait(self, *, path: str) -> None:
        self._thread_status_keys.pop(path, None)
        self._set_thread_status_nowait(path=path, status=None)

    def _update_thread_status_from_event(self, *, path: str, event: dict) -> None:
        del path
        del event

    async def _clear_all_thread_statuses(self) -> None:
        paths = {
            *self._thread_status_values.keys(),
            *self._thread_status_mode_values.keys(),
            *self._thread_status_keys.keys(),
        }
        for path in paths:
            await self.set_thread_status(path=path, status=None)
        self._thread_status_keys.clear()
        self._thread_status_values.clear()
        self._thread_status_mode_values.clear()
        self._thread_status_started_at_values.clear()
        self._thread_status_locks.clear()
        self._thread_status_generations.clear()

    async def _send_and_save_chat(
        self,
        thread_adapter: ThreadAdapter,
        to: RemoteParticipant,
        id: str,
        text: str,
        thread_attributes: dict,
    ):
        with tracer.start_as_current_span("chatbot.thread.message") as span:
            span.set_attributes(thread_attributes)
            span.set_attribute("role", "assistant")
            span.set_attribute(
                "from_participant_name",
                self.room.local_participant.get_attribute("name"),
            )
            span.set_attributes({"id": id, "text": text})
            await self.room.messaging.send_message(
                to=to,
                type="chat",
                message={"path": thread_adapter.path, "text": text},
            )
            thread_adapter.write_text_message(
                text=text,
                participant=self.room.local_participant,
            )

    async def _greet(
        self,
        *,
        thread_adapter: ThreadAdapter,
        thread_context: ChatThreadContext,
        participant: RemoteParticipant,
        thread_attributes: dict,
    ):
        if self._auto_greet_message is not None:
            thread_context.session.append_user_message(self._auto_greet_message)
            await self._send_and_save_chat(
                id=str(uuid.uuid4()),
                to=RemoteParticipant(id=participant.id),
                thread_adapter=thread_adapter,
                text=self._auto_greet_message,
                thread_attributes=thread_attributes,
            )

    def get_requirements(self):
        return [
            *super().get_requirements(),
        ]

    async def get_online_participants(
        self, *, thread: MeshDocument, exclude: Optional[list[Participant]] = None
    ):
        return get_online_participants(room=self._room, thread=thread, exclude=exclude)

    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return []

    async def get_thread_toolkits(
        self, *, thread_context: ChatThreadContext, participant: RemoteParticipant
    ) -> list[Toolkit]:
        toolkits = await self.get_required_toolkits(
            context=ToolContext(
                room=self.room,
                caller=self.room.local_participant,
                on_behalf_of=participant,
                caller_context=thread_context.to_caller_context(),
                event_handler=thread_context.emit,
            )
        )

        @tool(
            name="attach_file",
            description="attach a file to the thread so the user can see it",
        )
        async def attach_file(path: str):
            messages = thread_context.thread.root.get_elements_by_tag_name("messages")[
                0
            ]
            message = messages.append_child(
                tag_name="message",
                attributes={
                    "id": str(uuid.uuid4()),
                    "text": "",
                    "created_at": datetime.now(timezone.utc)
                    .isoformat()
                    .replace("+00:00", "Z"),
                    "author_name": self.room.local_participant.get_attribute("name"),
                },
            )
            message.append_child(tag_name="file", attributes={"path": path})

        toolkits.append(
            Toolkit(
                name="thread tools",
                tools=[attach_file],
            )
        )
        thread_list_tools = self._build_thread_list_tools()
        if len(thread_list_tools) > 0:
            toolkits.append(
                Toolkit(
                    name="chat thread list",
                    tools=thread_list_tools,
                )
            )

        toolkit = self._open_threads[thread_context.path].make_toolkit()

        return [*self._toolkits, *toolkits, toolkit]

    @abstractmethod
    def default_model(self) -> str: ...

    @abstractmethod
    async def create_thread_context(
        self,
        *,
        path: str,
        thread: MeshDocument,
        participants: list[RemoteParticipant],
        event_handler: Callable[[dict], None],
    ) -> ChatThreadContext: ...

    def create_thread_adapter(self, *, path: str) -> ThreadAdapter:
        if isinstance(self._llm_adapter, OpenAICompletionsAdapter):
            return CompletionsThreadAdapter(
                room=self.room,
                path=path,
                format_message=self.format_message,
            )

        return ResponsesThreadAdapter(
            room=self.room,
            path=path,
            format_message=self.format_message,
        )

    def _should_store_received_chat_message(self, *, message: dict[str, Any]) -> bool:
        store = message.get("store")
        return isinstance(store, bool) and store

    async def open_thread(self, *, path: str) -> ThreadAdapter:
        logger.info(f"opening thread {path}")
        if path not in self._open_threads:
            adapter = self.create_thread_adapter(path=path)
            await adapter.start()
            self._open_threads[path] = adapter

        return self._open_threads[path]

    async def close_thread(self, *, path: str) -> None:
        logger.info(f"closing thread {path}")

        adapter = self._open_threads.pop(path, None)
        if adapter is not None:
            await adapter.stop()

    async def on_thread_open(self, *, thread_context: ChatThreadContext):
        pass

    async def on_thread_clear(self, *, thread_context: ChatThreadContext):
        pass

    async def on_thread_cancel(self, *, thread_context: ChatThreadContext):
        pass

    async def on_thread_close(self, *, thread_context: ChatThreadContext):
        pass

    async def on_approved(
        self,
        *,
        thread_context: ChatThreadContext,
        from_participant: RemoteParticipant,
        message: dict,
    ):
        pass

    async def on_rejected(
        self,
        *,
        thread_context: ChatThreadContext,
        from_participant: RemoteParticipant,
        message: dict,
    ):
        pass

    async def on_thread_steer(
        self,
        *,
        thread_context: ChatThreadContext,
        from_participant: RemoteParticipant,
        message: dict,
    ) -> None:
        del thread_context
        del from_participant
        del message

    async def cancel_thread_task(
        self, *, path: str, thread_context: Optional[ChatThreadContext]
    ) -> None:
        del thread_context
        if path in self._thread_tasks:
            self._thread_tasks[path].cancel()

    async def _safe_invoke_thread_event(
        self,
        *,
        event_name: str,
        thread_context: ChatThreadContext,
        handler: Callable[..., Awaitable[None]],
    ) -> None:
        try:
            await handler(thread_context=thread_context)
        except ChanClosed:
            logger.debug(
                "chatbot thread event hook '%s' skipped because room channel is closed",
                event_name,
            )
        except Exception as e:
            logger.error(
                f"chatbot thread event hook '{event_name}' failed",
                exc_info=e,
            )

    async def _safe_invoke_chat_event(
        self,
        *,
        event_name: str,
        thread_context: ChatThreadContext,
        from_participant: RemoteParticipant,
        message: dict[str, Any],
        handler: Callable[..., Awaitable[None]],
    ) -> None:
        try:
            await handler(
                thread_context=thread_context,
                from_participant=from_participant,
                message=message,
            )
        except Exception as e:
            logger.error(
                f"chatbot chat event hook '{event_name}' failed",
                exc_info=e,
            )

    def get_thread_members(self, *, thread: MeshDocument) -> list[str]:
        results = []

        for prop in thread.root.get_children():
            if prop.tag_name == "members":
                for member in prop.get_children():
                    results.append(member.get_attribute("name"))

        return results

    async def get_rules(
        self, *, thread_context: ChatThreadContext, participant: RemoteParticipant
    ):
        rules = [*self._rules]

        if self._skill_dirs is not None and len(self._skill_dirs) > 0:
            rules.append(
                "You have access to to following skills which follow the agentskills spec:"
            )
            rules.append(await to_prompt([*(Path(p) for p in self._skill_dirs)]))
            rules.append(
                "Use the shell or storage tool to find out more about skills and execute them when they are required"
            )

        client = participant.get_attribute("client")

        if self._client_rules is not None and client is not None:
            cr = self._client_rules.get(client)
            if cr is not None:
                rules.extend(cr)

        # Without this rule 5.2 / 5.1 like to start their messages with things like "I could say"
        rules.append("based on the previous transcript, take your turn and respond")

        return rules

    async def on_chat_received(
        self,
        *,
        thread_context: ChatThreadContext,
        from_participant: RemoteParticipant,
        message: dict,
    ) -> Optional[str]:
        return None

    def format_message(self, *, user_name: str, message: str, iso_timestamp: str):
        return f"{user_name} said at {iso_timestamp}: {message}"

    def _chat_error_message(self, *, error: Exception) -> str:
        if isinstance(error, RoomException):
            message = str(error).strip()
            if message != "":
                return message
        if isinstance(error, aiohttp.WSServerHandshakeError):
            return OpenAIResponsesSessionContext._handshake_error_message(error)
        return "An unexpected error occured. Please try again later."

    async def _spawn_thread(self, path: str, messages: Chan[RoomMessage]):
        logger.debug("chatbot is starting a thread", extra={"path": path})
        opened = False

        current_file = None
        thread_context = None

        thread_attributes = None

        thread = None
        thread_adapter = None

        def handle_event(evt, participant: RemoteParticipant):
            if isinstance(evt, BaseModel):
                evt = evt.model_dump(mode="json")

            if evt.get("type") == "response.output_text.done" and thread is not None:
                online = get_online_participants(room=self._room, thread=thread)
                for online_participant in online:
                    if online_participant.id != self._room.local_participant.id:
                        logger.info(
                            f"replying to {online_participant.get_attribute('name')}"
                        )
                        self._room.messaging.send_message_nowait(
                            to=online_participant,
                            type="chat",
                            message={
                                "type": "chat",
                                "path": path,
                                "text": evt.get("text"),
                            },
                        )

                if participant not in online:
                    logger.info(f"replying to {participant.get_attribute('name')}")
                    self._room.messaging.send_message_nowait(
                        to=participant,
                        type="chat",
                        message={
                            "type": "chat",
                            "path": path,
                            "text": evt.get("text"),
                        },
                    )

            if thread_adapter is None:
                return

            thread_adapter.push(event=evt)
            self._update_thread_status_from_event(path=path, event=evt)

        async def _close_thread_context_session(
            context: ChatThreadContext | None,
        ) -> None:
            if context is None:
                return
            try:
                await context.__aexit__(None, None, None)
            except Exception as ex:
                logger.warning(
                    "unable to close chat session context for thread %s",
                    path,
                    exc_info=ex,
                )

        async def _activate_thread_context(
            context: ChatThreadContext,
        ) -> None:
            nonlocal thread_context

            await context.__aenter__()

            previous = self._thread_contexts.get(path)
            self._thread_contexts[path] = context
            thread_context = context

            if previous is not None and previous is not context:
                await _close_thread_context_session(previous)

        try:
            received = None

            while True:
                logger.debug(f"waiting for message on thread {path}")
                received = await messages.recv()

                logger.debug(f"received message on thread {path}: {received.type}")

                chat_with_participant = received.from_participant

                thread_attributes = {
                    "agent_name": self.name,
                    "agent_participant_id": self.room.local_participant.id,
                    "agent_participant_name": self.room.local_participant.get_attribute(
                        "name"
                    ),
                    "remote_participant_id": chat_with_participant.id,
                    "remote_participant_name": chat_with_participant.get_attribute(
                        "name"
                    ),
                    "path": path,
                }

                if current_file != chat_with_participant.get_attribute("current_file"):
                    logger.info(
                        f"{chat_with_participant.get_attribute('name')} is now looking at {chat_with_participant.get_attribute('current_file')}"
                    )
                    current_file = chat_with_participant.get_attribute("current_file")

                if thread is None:
                    with tracer.start_as_current_span("chatbot.thread.open") as span:
                        span.set_attributes(thread_attributes)

                        thread_adapter = await self.open_thread(path=path)
                        thread = thread_adapter.thread
                        if thread is None:
                            raise RoomException("thread was not opened")

                        thread_context = await self.create_thread_context(
                            path=path,
                            thread=thread,
                            participants=get_online_participants(
                                room=self.room, thread=thread
                            ),
                            event_handler=lambda evt: handle_event(
                                evt, chat_with_participant
                            ),
                        )
                        await _activate_thread_context(thread_context)

                        thread_adapter.append_messages(
                            context=thread_context.session,
                        )

                if received.type == "opened":
                    if not opened:
                        opened = True

                        if thread_context is not None:
                            await self._safe_invoke_thread_event(
                                event_name="open",
                                thread_context=thread_context,
                                handler=self.on_thread_open,
                            )

                        await self._greet(
                            thread_context=thread_context,
                            participant=chat_with_participant,
                            thread_adapter=thread_adapter,
                            thread_attributes=thread_attributes,
                        )
                elif received.type == "clear":
                    thread_context = await self.create_thread_context(
                        path=path,
                        thread=thread,
                        participants=get_online_participants(
                            room=self.room, thread=thread
                        ),
                        event_handler=lambda evt: handle_event(
                            evt, chat_with_participant
                        ),
                    )
                    await _activate_thread_context(thread_context)
                    messages_element: Element = thread.root.get_children_by_tag_name(
                        "messages"
                    )[0]
                    for child in list(messages_element.get_children()):
                        child.delete()

                    if thread_context is not None:
                        await self._safe_invoke_thread_event(
                            event_name="clear",
                            thread_context=thread_context,
                            handler=self.on_thread_clear,
                        )

                elif received.type == "chat":
                    if thread is None:
                        logger.info("thread is not open", extra={"path": path})
                        break

                    logger.debug(
                        "chatbot received a chat",
                        extra={
                            "context": (
                                thread_context.context_id
                                if thread_context is not None
                                else None
                            ),
                            "participant_id": self.room.local_participant.id,
                            "participant_name": self.room.local_participant.get_attribute(
                                "name"
                            ),
                            "text": received.message["text"],
                        },
                    )

                if received is not None and received.type == "chat":
                    with tracer.start_as_current_span("chatbot.thread.message") as span:
                        span.set_attributes(thread_attributes)
                        span.set_attribute("role", "user")
                        span.set_attribute(
                            "from_participant_name",
                            chat_with_participant.get_attribute("name"),
                        )

                        raw_attachments = received.message.get("attachments")
                        attachments = (
                            raw_attachments if isinstance(raw_attachments, list) else []
                        )
                        span.set_attribute("attachments", json.dumps(attachments))

                        text = received.message["text"]
                        span.set_attributes({"text": text})

                        if (
                            thread_adapter is not None
                            and self._should_store_received_chat_message(
                                message=received.message
                            )
                        ):
                            thread_adapter.write_text_message(
                                text=text,
                                participant=chat_with_participant,
                                attachments=attachments,
                            )

                        try:
                            if thread_context is None:
                                thread_context = await self.create_thread_context(
                                    path=path,
                                    thread=thread,
                                    participants=get_online_participants(
                                        room=self.room, thread=thread
                                    ),
                                    event_handler=lambda evt: handle_event(
                                        evt, chat_with_participant
                                    ),
                                )
                                await _activate_thread_context(thread_context)
                            else:
                                thread_context.participants = get_online_participants(
                                    room=self.room, thread=thread
                                )

                            await self.set_thread_status(
                                path=path,
                                status="Thinking",
                                mode=self.processing_thread_status_mode(
                                    path=path,
                                    thread_context=thread_context,
                                ),
                            )

                            result = await self.on_chat_received(
                                thread_context=thread_context,
                                from_participant=chat_with_participant,
                                message=received.message,
                            )
                            received.result.set_result(result)

                        except Exception as e:
                            handled_error: Exception = e
                            if isinstance(e, aiohttp.WSServerHandshakeError):
                                handled_error = RoomException(
                                    OpenAIResponsesSessionContext._handshake_error_message(
                                        e
                                    ),
                                    status_code=e.status,
                                )

                            if isinstance(handled_error, RoomException):
                                logger.warning(
                                    "A room error was encountered", exc_info=e
                                )
                            else:
                                logger.error("An error was encountered", exc_info=e)

                            text = self._chat_error_message(error=handled_error)
                            await self._send_and_save_chat(
                                thread_adapter=thread_adapter,
                                to=chat_with_participant,
                                id=str(uuid.uuid4()),
                                text=text,
                                thread_attributes=thread_attributes,
                            )
                            received.result.set_result(text)

                        finally:
                            await self.clear_thread_status(path=path)

        finally:

            async def cleanup():
                if self.room is not None:
                    logger.info(f"thread was ended {path}")
                    logger.info("chatbot thread ended", extra={"path": path})

                    thread_context = self._thread_contexts.pop(path, None)
                    if thread_context is not None:
                        await self._safe_invoke_thread_event(
                            event_name="close",
                            thread_context=thread_context,
                            handler=self.on_thread_close,
                        )
                        await _close_thread_context_session(thread_context)

                    await self.close_thread(path=path)

            await asyncio.shield(cleanup())

    def _get_message_channel(self, key: str) -> Chan[_QueuedChatMessage]:
        if key not in self._message_channels:
            chan = Chan[_QueuedChatMessage]()
            self._message_channels[key] = chan

        chan = self._message_channels[key]

        return chan

    async def stop(self):
        room = self._room
        message_channels = list(self._message_channels.values())
        for channel in message_channels:
            channel.close()

        thread_tasks = list(self._thread_tasks.values())
        for thread in thread_tasks:
            thread.cancel()

        if len(thread_tasks) > 0:
            await asyncio.gather(*thread_tasks, return_exceptions=True)

        self._thread_tasks.clear()
        self._message_channels.clear()

        await self._clear_all_thread_statuses()
        await self._close_thread_list_document(room=room)
        await super().stop()

    async def _on_get_thread_toolkits_message(self, *, message: RoomMessage):
        path = message.message["path"]

        chat_with_participant = None
        for participant in self._room.messaging.get_participants():
            if participant.id == message.from_participant_id:
                chat_with_participant = participant
                break

        if chat_with_participant is None:
            logger.warning(
                "participant does not have messaging enabled, skipping message"
            )
            return

        tool_providers = self.get_toolkit_builders()
        self._room.messaging.send_message_nowait(
            to=chat_with_participant,
            type="set_thread_tool_providers",
            message={
                "path": path,
                "tool_providers": [{"name": t.name} for t in tool_providers],
            },
        )

    async def _queue_chat_message(
        self,
        *,
        context: ToolContext,
        path: str,
        message: dict[str, Any],
        wait_for_result: bool,
        store: bool = True,
    ) -> Optional[str]:
        text_value = message.get("text")
        text = text_value if isinstance(text_value, str) else ""
        attachments: list[dict[str, str]] = []
        raw_attachments = message.get("attachments")
        if isinstance(raw_attachments, list):
            for attachment in raw_attachments:
                if not isinstance(attachment, dict):
                    continue
                path_value = attachment.get("path")
                if not isinstance(path_value, str):
                    continue
                normalized_path = path_value.strip()
                if normalized_path == "":
                    continue
                attachments.append({"path": normalized_path})

        payload = {
            **message,
            "path": path,
            "text": text,
            "attachments": attachments,
            "store": store,
        }
        qm = _QueuedChatMessage(
            type="chat",
            message=payload,
            from_participant=context.on_behalf_of or context.caller,
        )

        messages = self._ensure_thread(path=path)
        messages.send_nowait(qm)

        if wait_for_result:
            return await qm.result
        return None

    async def _seed_new_thread_members(
        self,
        *,
        path: str,
        members: list[Participant | str],
    ) -> None:
        thread = await self.room.sync.open(path=path, schema=thread_schema)
        try:
            members_elements = thread.root.get_children_by_tag_name("members")
            members_element = (
                members_elements[0]
                if len(members_elements) > 0
                else thread.root.append_child(tag_name="members")
            )

            existing_members: set[str] = set()
            for child in members_element.get_children():
                if child.tag_name != "member":
                    continue
                member_name = child.get_attribute("name")
                if not isinstance(member_name, str):
                    continue
                normalized_name = member_name.strip()
                if normalized_name == "":
                    continue
                existing_members.add(normalized_name)

            for participant in members:
                participant_name = (
                    participant.get_attribute("name")
                    if isinstance(participant, Participant)
                    else participant
                )
                if not isinstance(participant_name, str):
                    continue

                normalized_name = participant_name.strip()
                if normalized_name == "" or normalized_name in existing_members:
                    continue

                members_element.append_child(
                    tag_name="member",
                    attributes={"name": normalized_name},
                )
                existing_members.add(normalized_name)
        finally:
            await self.room.sync.close(path=path)

    def _build_thread_list_tools(self) -> list[FunctionTool]:
        if self._thread_list_dir() is None:
            return []

        read_file_hint = (
            "Use read_file with a thread path to read that thread's contents."
        )
        outer = self

        def to_json_entry(entry: Element) -> dict[str, str]:
            return {
                "name": str(entry.get_attribute("name") or ""),
                "path": str(entry.get_attribute("path") or ""),
                "modified_at": str(entry.get_attribute("modified_at") or ""),
                "created_at": str(entry.get_attribute("created_at") or ""),
            }

        @tool(
            name="list_threads",
            description="lists recent threads sorted by last modified date (newest first). Use read_file with a thread path to read that thread's contents.",
        )
        def list_threads(*, limit: int = 20, offset: int = 0) -> JsonContent:
            normalized_offset = max(0, int(offset))
            normalized_limit = max(1, min(200, int(limit)))
            if outer._thread_list_dir() is None:
                return JsonContent(
                    json={
                        "threads": [],
                        "total": 0,
                        "offset": normalized_offset,
                        "limit": normalized_limit,
                        "message": "thread list is not enabled for this chatbot",
                        "read_file_hint": read_file_hint,
                    }
                )

            entries = outer._sorted_thread_list_entries()
            if len(entries) == 0:
                return JsonContent(
                    json={
                        "threads": [],
                        "total": 0,
                        "offset": normalized_offset,
                        "limit": normalized_limit,
                        "message": "no threads were found in the thread list",
                        "read_file_hint": read_file_hint,
                    }
                )

            selected = outer._thread_list_slice(
                entries=entries,
                limit=limit,
                offset=offset,
            )
            if len(selected) == 0:
                return JsonContent(
                    json={
                        "threads": [],
                        "total": len(entries),
                        "offset": normalized_offset,
                        "limit": normalized_limit,
                        "message": "no threads were found for the requested limit/offset",
                        "read_file_hint": read_file_hint,
                    }
                )

            return JsonContent(
                json={
                    "threads": [to_json_entry(entry) for entry in selected],
                    "total": len(entries),
                    "offset": normalized_offset,
                    "limit": normalized_limit,
                    "sort": "modified_at_desc",
                    "read_file_hint": read_file_hint,
                }
            )

        @tool(
            name="grep_thread_list",
            description="searches the thread list for matching thread names and paths. Use read_file with a thread path to read that thread's contents.",
        )
        def grep_thread_list(*, pattern: str, ignore_case: bool = True) -> JsonContent:
            if outer._thread_list_dir() is None:
                return JsonContent(
                    json={
                        "threads": [],
                        "total_matches": 0,
                        "pattern": pattern,
                        "ignore_case": ignore_case,
                        "message": "thread list is not enabled for this chatbot",
                        "read_file_hint": read_file_hint,
                    }
                )

            needle = pattern.strip()
            if needle == "":
                return JsonContent(
                    json={
                        "threads": [],
                        "total_matches": 0,
                        "pattern": needle,
                        "ignore_case": ignore_case,
                        "message": "pattern is required",
                        "read_file_hint": read_file_hint,
                    }
                )

            flags = re.IGNORECASE if ignore_case else 0
            try:
                matcher = re.compile(needle, flags)
            except re.error as ex:
                return JsonContent(
                    json={
                        "threads": [],
                        "total_matches": 0,
                        "pattern": needle,
                        "ignore_case": ignore_case,
                        "error": "invalid_regex_pattern",
                        "message": f"invalid regex pattern: {ex}",
                        "read_file_hint": read_file_hint,
                    }
                )

            matches: list[dict[str, str]] = []
            for entry in outer._sorted_thread_list_entries():
                name = entry.get_attribute("name")
                path = entry.get_attribute("path")
                created_at = entry.get_attribute("created_at")
                modified_at = entry.get_attribute("modified_at")
                haystack = f"{name}\n{path}\n{created_at}\n{modified_at}"
                if matcher.search(haystack) is None:
                    continue
                matches.append(to_json_entry(entry))

            if len(matches) == 0:
                return JsonContent(
                    json={
                        "threads": [],
                        "total_matches": 0,
                        "pattern": needle,
                        "ignore_case": ignore_case,
                        "message": "no matching threads were found",
                        "read_file_hint": read_file_hint,
                    }
                )

            return JsonContent(
                json={
                    "threads": matches,
                    "total_matches": len(matches),
                    "pattern": needle,
                    "ignore_case": ignore_case,
                    "read_file_hint": read_file_hint,
                }
            )

        return [list_threads, grep_thread_list]

    async def get_exposed_toolkits(self) -> list[RemoteToolkit]:
        exposed_toolkits = await super().get_exposed_toolkits()

        @tool(
            description=f"sends a chat to {self.room.local_participant.get_attribute('name')} and gets the response"
        )
        async def ask(context: ToolContext, *, path: str, text: str) -> str:
            response = await self._queue_chat_message(
                context=context,
                path=path,
                message={"text": text},
                wait_for_result=True,
            )
            if response is None:
                raise RoomException("chat response was empty")
            return response

        tools_schema = {
            "type": "object",
            "additionalProperties": False,
            "required": ["message"],
            "properties": {
                "message": {
                    "type": "object",
                    "additionalProperties": False,
                    "required": ["text"],
                    "properties": {
                        "text": {"type": "string"},
                        "attachments": {
                            "anyOf": [
                                {
                                    "type": "array",
                                    "items": {"$ref": "#/$defs/ChatAttachment"},
                                },
                                {"type": "null"},
                            ]
                        },
                    },
                }
            },
            "$defs": {
                "ChatAttachment": {
                    "type": "object",
                    "additionalProperties": False,
                    "required": ["path"],
                    "properties": {
                        "path": {"type": "string"},
                    },
                }
            },
        }

        message_tools_schema, defs = build_tools_property_schema(
            toolkit_builders=self.get_toolkit_builders()
        )
        if message_tools_schema is not None:
            tools_schema["properties"]["message"]["properties"]["tools"] = (
                message_tools_schema
            )
        else:
            # Backward compatibility: some clients always send `tools: []`.
            tools_schema["properties"]["message"]["properties"]["tools"] = {
                "anyOf": [
                    {
                        "type": "array",
                        "items": ensure_strict_json_schema({}),
                    },
                    {"type": "null"},
                ],
            }

        if len(defs) > 0:
            for key, value in defs.items():
                tools_schema["$defs"][key] = value

        tools_schema = ensure_strict_json_schema(tools_schema)

        outer = self

        class NewThreadTool(FunctionTool):
            def __init__(self):
                super().__init__(
                    name="new_thread",
                    description=f"creates a new thread for {outer.room.local_participant.get_attribute('name')} and sends the first chat message, returning the new thread path and name",
                    input_schema=tools_schema,
                    supports_context=True,
                )

            async def execute(
                self,
                context: ToolContext,
                *,
                message: dict[str, Any],
            ) -> JsonContent:
                text_value = message.get("text")
                text = text_value if isinstance(text_value, str) else ""
                payload = {**message, "text": text}

                path, friendly_name = await outer._generate_new_thread_info(
                    context=context,
                    text=text,
                )
                await outer._seed_new_thread_members(
                    path=path,
                    members=[
                        outer.room.local_participant,
                        context.on_behalf_of or context.caller,
                    ],
                )
                outer._record_new_thread_in_index(path=path, name=friendly_name)
                await outer._queue_chat_message(
                    context=context,
                    path=path,
                    message=payload,
                    wait_for_result=False,
                )
                return JsonContent(json={"path": path, "name": friendly_name})

        tools: list[FunctionTool] = [
            ask,
            NewThreadTool(),
            *self._build_thread_list_tools(),
        ]

        chatbot_toolkit = RemoteToolkit(
            name="chat",
            description=f"tools for interacting with {self.name}",
            public=False,
            tools=tools,
            validation_mode="content_types",
        )

        exposed_toolkits.append(chatbot_toolkit)
        return exposed_toolkits

    def _ensure_thread(self, path: str) -> Chan[_QueuedChatMessage]:
        messages = self._get_message_channel(path)
        if path not in self._thread_tasks or self._thread_tasks[path].done():

            def thread_done(task: asyncio.Task):
                self._thread_tasks.pop(path, None)
                self._message_channels.pop(path, None)
                try:
                    task.result()
                except CancelledError:
                    pass
                except Exception as e:
                    logger.error(f"The chat thread ended with an error {e}", exc_info=e)

            logger.debug(f"spawning chat thread for {path}")
            task = asyncio.create_task(self._spawn_thread(messages=messages, path=path))
            task.add_done_callback(thread_done)

            self._thread_tasks[path] = task

        return messages

    def on_message(
        self, message: RoomMessage, from_participant: Optional[RemoteParticipant] = None
    ) -> _QueuedChatMessage | None:
        if message.type == "get_thread_toolkit_builders":
            task = asyncio.create_task(
                self._on_get_thread_toolkits_message(message=message)
            )

            def on_done(task: asyncio.Task):
                try:
                    task.result()
                except Exception as ex:
                    logger.error(f"unable to get tool providers {ex}", exc_info=ex)

            task.add_done_callback(on_done)

        if (
            message.type == "chat"
            or message.type == "opened"
            or message.type == "clear"
        ):
            path = message.message["path"]

            logger.debug(f"queued incoming message for thread {path}: {message.type}")

            if from_participant is None:
                for participant in self._room.messaging.get_participants():
                    if participant.id == message.from_participant_id:
                        from_participant = participant
                        break

                if from_participant is None:
                    logger.warning(
                        "participant does not have messaging enabled, skipping message"
                    )
                    return

            msg = _QueuedChatMessage(
                type=message.type,
                message=message.message,
                from_participant=from_participant,
            )

            messages = self._ensure_thread(path=path)

            messages.send_nowait(msg)

            return msg

        elif message.type == "steer":
            path = message.message["path"]

            if from_participant is None:
                for participant in self._room.messaging.get_participants():
                    if participant.id == message.from_participant_id:
                        from_participant = participant
                        break

                if from_participant is None:
                    logger.warning(
                        "participant does not have messaging enabled, skipping message"
                    )
                    return

            async def handle_steer():
                thread_context = self._thread_contexts.get(path)
                if thread_context is None:
                    logger.warning(
                        f"unable to process steer message for thread {path}: thread is not open"
                    )
                    return

                await self._safe_invoke_chat_event(
                    event_name="steer",
                    thread_context=thread_context,
                    from_participant=from_participant,
                    message=message.message,
                    handler=self.on_thread_steer,
                )

            task = asyncio.create_task(handle_steer())

            def on_done(task: asyncio.Task):
                try:
                    task.result()
                except CancelledError:
                    pass
                except Exception as ex:
                    logger.error(
                        f"unable to process steer message for thread {path}",
                        exc_info=ex,
                    )

            task.add_done_callback(on_done)

        elif message.type == "approved" or message.type == "rejected":
            path = message.message["path"]

            if from_participant is None:
                for participant in self._room.messaging.get_participants():
                    if participant.id == message.from_participant_id:
                        from_participant = participant
                        break

                if from_participant is None:
                    logger.warning(
                        "participant does not have messaging enabled, skipping message"
                    )
                    return

            async def handle_approval_response():
                thread_context = self._thread_contexts.get(path)
                if thread_context is None:
                    logger.warning(
                        f"unable to process {message.type} for thread {path}: thread is not open"
                    )
                    return

                if message.type == "approved":
                    await self._safe_invoke_chat_event(
                        event_name="approved",
                        thread_context=thread_context,
                        from_participant=from_participant,
                        message=message.message,
                        handler=self.on_approved,
                    )
                else:
                    await self._safe_invoke_chat_event(
                        event_name="rejected",
                        thread_context=thread_context,
                        from_participant=from_participant,
                        message=message.message,
                        handler=self.on_rejected,
                    )

            task = asyncio.create_task(handle_approval_response())

            def on_done(task: asyncio.Task):
                try:
                    task.result()
                except CancelledError:
                    pass
                except Exception as ex:
                    logger.error(
                        f"unable to process {message.type} message for thread {path}",
                        exc_info=ex,
                    )

            task.add_done_callback(on_done)

        elif message.type == "cancel":
            path = message.message["path"]

            async def handle_cancel():
                thread_context = self._thread_contexts.get(path)
                if thread_context is not None:
                    await self._safe_invoke_thread_event(
                        event_name="cancel",
                        thread_context=thread_context,
                        handler=self.on_thread_cancel,
                    )

                await self.cancel_thread_task(
                    path=path,
                    thread_context=thread_context,
                )

            task = asyncio.create_task(handle_cancel())

            def on_done(task: asyncio.Task):
                try:
                    task.result()
                except CancelledError:
                    pass
                except Exception as ex:
                    logger.error(
                        f"unable to process cancel message for thread {path}",
                        exc_info=ex,
                    )

            task.add_done_callback(on_done)

        elif message.type == "typing":

            def callback(task: asyncio.Task):
                try:
                    task.result()
                except CancelledError:
                    pass
                except Exception:
                    pass

            async def remove_timeout(id: str):
                await asyncio.sleep(1)
                self._is_typing.pop(id)

            if message.from_participant_id in self._is_typing:
                self._is_typing[message.from_participant_id].cancel()

            timeout = asyncio.create_task(
                remove_timeout(id=message.from_participant_id)
            )
            timeout.add_done_callback(callback)

            self._is_typing[message.from_participant_id] = timeout

        return None

    async def start(self, *, room):
        await super().start(room=room)

        logger.debug("Starting chatbot")

        if self._threading_mode is not None:
            await self.room.local_participant.set_attribute(
                "meshagent.chatbot.threading", self._threading_mode
            )

        thread_dir = self._thread_list_dir()
        if thread_dir is not None:
            await self.room.local_participant.set_attribute(
                "meshagent.chatbot.thread-dir", thread_dir
            )

        thread_list_path = self._thread_list_index_path()
        if thread_list_path is not None:
            await self.room.local_participant.set_attribute(
                "meshagent.chatbot.thread-list", thread_list_path
            )

        await self.room.local_participant.set_attribute(
            "empty_state_title", self._empty_state_title
        )

        await self._open_thread_list_document()

        room.messaging.on("message", self.on_message)

        if self._auto_greet_message is not None:

            def on_participant_added(participant: RemoteParticipant):
                # will spawn the initial thread
                self._get_message_channel(participant.id)

            room.messaging.on("participant_added", on_participant_added)

        logger.debug("Enabling chatbot messaging")
        await room.messaging.enable()


class ChatBot(ChatBotBase):
    def __init__(
        self,
        *,
        name=None,
        title=None,
        description=None,
        requires: Optional[list[Requirement]] = None,
        llm_adapter: LLMAdapter,
        toolkits: Optional[list[Toolkit]] = None,
        rules: Optional[list[str]] = None,
        client_rules: Optional[dict[str, list[str]]] = None,
        auto_greet_message: Optional[str] = None,
        empty_state_title: Optional[str] = None,
        annotations: Optional[list[str]] = None,
        decision_model: Optional[str] = None,
        decision_options: Optional[dict] = None,
        always_reply: Optional[bool] = None,
        skill_dirs: Optional[list[str]] = None,
        thread_dir: Optional[str] = None,
        threading_mode: Optional[str] = None,
        thread_name_rules: Optional[list[str]] = None,
    ):
        self._llm_adapter = llm_adapter
        if decision_model is None:
            decision_model = "gpt-5.4"
            decision_options = {"reasoning": {"effort": "none"}}

        self._decision_model = decision_model
        self._decision_options = decision_options

        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            toolkits=toolkits,
            rules=rules,
            client_rules=client_rules,
            auto_greet_message=auto_greet_message,
            empty_state_title=empty_state_title,
            annotations=annotations,
            always_reply=always_reply,
            skill_dirs=skill_dirs,
            thread_dir=thread_dir,
            threading_mode=threading_mode,
            thread_name_rules=thread_name_rules,
        )

    def default_model(self) -> str:
        return self._llm_adapter.default_model()

    def thread_name_adapter(self) -> Optional[LLMAdapter]:
        return self._llm_adapter

    def _status_event_details(
        self, *, event: dict
    ) -> tuple[Optional[str], Optional[str], Optional[str]]:
        normalized_event = None
        event_type = event.get("type")
        if event_type in ("agent.event", "codex.event"):
            normalized_event = event
        else:
            normalized_event = response_event_to_agent_event(event)

        if not isinstance(normalized_event, dict):
            return None, None, None

        kind = normalized_event.get("kind")
        if not isinstance(kind, str):
            kind = ""
        kind = kind.strip().lower()
        if kind not in (
            "exec",
            "tool",
            "web",
            "search",
            "diff",
            "image",
        ):
            return None, None, None

        state = normalized_event.get("state")
        if not isinstance(state, str):
            state = ""
        state = state.strip().lower()

        key = None
        for candidate in (
            normalized_event.get("correlation_key"),
            normalized_event.get("event_key"),
            normalized_event.get("item_id"),
            normalized_event.get("name"),
            normalized_event.get("method"),
        ):
            if isinstance(candidate, str) and candidate.strip() != "":
                key = candidate.strip()
                break

        text = None
        for candidate in (
            normalized_event.get("headline"),
            normalized_event.get("summary"),
            normalized_event.get("name"),
            normalized_event.get("method"),
        ):
            if isinstance(candidate, str):
                normalized = candidate.strip()
                if normalized != "":
                    text = normalized
                    break

        return key, state, text

    def _update_thread_status_from_event(self, *, path: str, event: dict) -> None:
        key, state, text = self._status_event_details(event=event)
        if state is None:
            return

        is_active = state in (
            "queued",
            "in_progress",
            "running",
            "pending",
            "searching",
        )
        if is_active:
            if text is None:
                return
            if key is not None:
                self._thread_status_keys[path] = key
            self._set_thread_status_nowait(path=path, status=text)
            return

        if key is not None:
            tracked = self._thread_status_keys.get(path)
            if tracked is not None and tracked == key:
                self._thread_status_keys.pop(path, None)
                self._set_thread_status_nowait(path=path, status="Thinking")
            return

        if state in ("completed", "failed", "cancelled"):
            self._set_thread_status_nowait(path=path, status="Thinking")

    async def create_thread_context(
        self,
        *,
        path: str,
        thread: MeshDocument,
        participants: list[RemoteParticipant],
        event_handler: Callable[[dict], None],
    ) -> ChatThreadContext:
        return ChatThreadContext(
            path=path,
            thread=thread,
            participants=participants,
            event_handler=event_handler,
            session=await self.init_session(),
        )

    def _create_default_session(self) -> AgentSessionContext:
        context = self._llm_adapter.create_session()
        context.append_rules(self._rules)
        return context

    async def init_session(self) -> AgentSessionContext:
        legacy_initializer = type(self).init_chat_context
        if legacy_initializer is not ChatBot.init_chat_context:
            cls = type(self)
            if cls not in _legacy_chatbot_init_chat_context_warned:
                warnings.warn(
                    (
                        f"{cls.__name__}.init_chat_context() is deprecated and will be removed in a future release. "
                        "Override init_session() instead."
                    ),
                    DeprecationWarning,
                    stacklevel=2,
                )
                _legacy_chatbot_init_chat_context_warned.add(cls)
            return await legacy_initializer(self)
        return self._create_default_session()

    # Backwards compatibility for existing subclasses overriding init_chat_context.
    async def init_chat_context(self) -> AgentSessionContext:
        warnings.warn(
            "init_chat_context() is deprecated and will be removed in a future release. Use init_session() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._create_default_session()

    async def prepare_llm_context(self, *, thread_context: ChatThreadContext):
        """
        called prior to sending the request to the LLM in case the agent needs to modify the context prior to sending
        """
        pass

    def prepare_chat_context(self, *, chat_context: AgentSessionContext):
        pass

    async def get_thread_toolkits(
        self, *, thread_context: ChatThreadContext, participant: RemoteParticipant
    ) -> list[Toolkit]:
        toolkits = await super().get_thread_toolkits(
            thread_context=thread_context, participant=participant
        )

        if isinstance(self._llm_adapter, OpenAIResponsesAdapter):
            toolkits.insert(
                0,
                Toolkit(
                    name="reasoning",
                    tools=[
                        ChatBotReasoningTool(
                            room=self._room,
                            thread_context=thread_context,
                        )
                    ],
                ),
            )

        return toolkits

    async def should_reply(
        self,
        *,
        context: ChatThreadContext,
        has_more_than_one_other_user: bool,
        toolkits: list[Toolkit],
        from_user: RemoteParticipant,
        online: list[Participant],
    ) -> bool:
        if not has_more_than_one_other_user or self._always_reply:
            return True

        online_set = {}

        all_members = []
        online_members = []

        for m in self.get_thread_members(thread=context.thread):
            all_members.append(m)

        for o in online:
            if o.get_attribute("name") not in online_set:
                online_set[o.get_attribute("name")] = True
                online_members.append(o.get_attribute("name"))

        logger.info(
            "multiple participants detected, checking whether agent should reply to conversation"
        )

        toolkits_json = ""
        for toolkit in toolkits:
            toolkits_json = (
                toolkits_json
                + f"\n{toolkit.name} ({toolkit.title}): {toolkit.description}"
            )
            for t in toolkit.tools:
                toolkits_json = (
                    toolkits_json + f"\n - {t.name} ({t.title}): {t.description}"
                )

        cloned_context = context.session.copy()
        async with cloned_context:
            cloned_context.replace_rules(
                rules=[
                    "examine the conversation so far and return whether the user is expecting a reply from you or another user as the next message in the conversation",
                    f'your name (the assistant) is "{self.room.local_participant.get_attribute("name")}"',
                    "if the user mentions a person with another name, they aren't talking to you unless they also mention you",
                    "if the user poses a question to everyone, they are talking to you",
                    "to help identify the different users in the conversation, every message in the thread will start with '{user_name} said at {time}'",
                    f"members of thread are currently {all_members}",
                    f"users online currently are {online_members}",
                    "if in doubt, reply to the user",
                    f"if the user is asking for something that these toolkits can do, they want an answer from you: {toolkits_json}",
                    "if the user they appear to be talking to is offline, then they probably are talking to you",
                ]
            )
            response = await self._llm_adapter.next(
                context=cloned_context,
                room=self._room,
                model=self._decision_model or self._llm_adapter.default_model(),
                options=self._decision_options,
                on_behalf_of=from_user,
                toolkits=[],
                output_schema={
                    "type": "object",
                    "required": ["reasoning", "expecting_assistant_reply", "next_user"],
                    "additionalProperties": False,
                    "properties": {
                        "reasoning": {
                            "type": "string",
                            "description": "explain why you think the user was or was not expecting you to reply",
                        },
                        "next_user": {
                            "type": "string",
                            "description": "who would be expected to send the next message in the conversation",
                        },
                        "expecting_assistant_reply": {"type": "boolean"},
                    },
                },
            )

        logger.info(f"should reply check returned {response}")

        parsed_response = response
        if isinstance(parsed_response, str):
            try:
                parsed_response = json.loads(parsed_response)
            except json.JSONDecodeError:
                logger.warning(
                    "should reply check returned unstructured text, defaulting to reply"
                )
                return True

        if not isinstance(parsed_response, dict):
            logger.warning(
                "should reply check returned %s, defaulting to reply",
                type(parsed_response).__name__,
            )
            return True

        expecting_assistant_reply = parsed_response.get("expecting_assistant_reply")
        if isinstance(expecting_assistant_reply, bool):
            return expecting_assistant_reply

        logger.warning(
            "should reply check returned payload without a boolean expecting_assistant_reply, defaulting to reply"
        )
        return True

    async def on_thread_open(self, *, thread_context: ChatThreadContext):
        await self.clear_thread_status(path=thread_context.path)

    async def on_thread_clear(self, *, thread_context: ChatThreadContext):
        await self.clear_thread_status(path=thread_context.path)

    async def on_thread_cancel(self, *, thread_context: ChatThreadContext):
        await self.clear_thread_status(path=thread_context.path)

    async def on_thread_close(self, *, thread_context: ChatThreadContext):
        await self.clear_thread_status(path=thread_context.path)

    async def on_chat_received(
        self,
        *,
        thread_context: ChatThreadContext,
        from_participant: RemoteParticipant,
        message: dict,
    ) -> Optional[str]:
        rules = await self.get_rules(
            thread_context=thread_context,
            participant=from_participant,
        )
        thread_context.session.replace_rules(rules)
        self._append_current_file_context(
            thread_context=thread_context,
            participant=from_participant,
        )

        attachments = message.get("attachments", [])
        for attachment in attachments:
            thread_context.session.append_assistant_message(
                message=f"the user attached a file at the path '{attachment['path']}'"
            )

        text = message["text"]
        iso_timestamp = datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")
        thread_context.session.append_user_message(
            message=self.format_message(
                user_name=from_participant.get_attribute("name"),
                message=text,
                iso_timestamp=iso_timestamp,
            )
        )

        model = message.get("model", self.default_model())

        with tracer.start_as_current_span("chatbot.llm"):
            with tracer.start_as_current_span("get_thread_toolkits"):
                thread_toolkits = await self.get_thread_toolkits(
                    thread_context=thread_context,
                    participant=from_participant,
                )

            with tracer.start_as_current_span("get_thread_toolkit_builders"):
                thread_tool_providers = self.get_toolkit_builders()

            await self.prepare_llm_context(thread_context=thread_context)

            message_toolkits = [*thread_toolkits]
            message_tools = message.get("tools")

            if message_tools is not None and len(message_tools) > 0:
                message_toolkits.extend(
                    await make_toolkits(
                        room=self.room,
                        model=model,
                        providers=thread_tool_providers,
                        tools=message_tools,
                    )
                )

        online = await self.get_online_participants(
            thread=thread_context.thread, exclude=[self.room.local_participant]
        )

        for participant in get_online_participants(
            room=self._room, thread=thread_context.thread
        ):
            self._room.messaging.send_message_nowait(
                to=participant,
                type="listening",
                message={"listening": True, "path": thread_context.path},
            )

        has_more_than_one_other_user = False

        thread_participants = []

        for member_name in self.get_thread_members(thread=thread_context.thread):
            thread_participants.append(member_name)
            if member_name != self._room.local_participant.get_attribute(
                "name"
            ) and member_name != from_participant.get_attribute("name"):
                has_more_than_one_other_user = True
                break

        thread_context.session.metadata["thread_participants"] = thread_participants

        reply = await self.should_reply(
            has_more_than_one_other_user=has_more_than_one_other_user,
            online=online,
            context=thread_context,
            toolkits=message_toolkits,
            from_user=from_participant,
        )

        for participant in get_online_participants(
            room=self._room, thread=thread_context.thread
        ):
            self._room.messaging.send_message_nowait(
                to=participant,
                type="listening",
                message={"listening": False, "path": thread_context.path},
            )

        if not reply:
            return None

        self.prepare_chat_context(chat_context=thread_context.session)
        self._touch_thread_in_index(path=thread_context.path)

        return await self._llm_adapter.next(
            context=thread_context.session,
            room=self._room,
            toolkits=message_toolkits,
            event_handler=thread_context.emit,
            model=model,
            on_behalf_of=from_participant,
        )

    def _normalize_current_file(self, *, value: object) -> str | None:
        if not isinstance(value, str):
            return None
        trimmed = value.strip()
        if trimmed == "":
            return None
        return trimmed

    def _participant_name_for_context(self, *, participant: Participant) -> str:
        name = participant.get_attribute("name")
        if isinstance(name, str) and name.strip() != "":
            return name
        return "the user"

    def _append_current_file_context(
        self,
        *,
        thread_context: ChatThreadContext,
        participant: Participant,
    ) -> None:
        metadata = thread_context.session.metadata
        state = metadata.get("current_file_by_participant")
        if not isinstance(state, dict):
            state = {}
            metadata["current_file_by_participant"] = state

        current_file = self._normalize_current_file(
            value=participant.get_attribute("current_file")
        )
        participant_key = participant.id
        had_previous = participant_key in state
        previous_file = state.get(participant_key)

        if had_previous and previous_file == current_file:
            return
        if not had_previous and current_file is None:
            return

        participant_name = self._participant_name_for_context(participant=participant)
        if current_file is None:
            thread_context.session.append_assistant_message(
                message=f"{participant_name} is not currently viewing any files."
            )
        else:
            thread_context.session.append_assistant_message(
                message=(
                    f"{participant_name} is currently viewing the file at "
                    f"the path: {current_file}"
                )
            )

        state[participant_key] = current_file



import asyncio
import json
import uuid
from datetime import datetime, timezone
from typing import Any, Optional

from meshagent.api import Element, RoomException

from .thread_adapter import ThreadAdapter, tracer

_ACTIVE_STATES = {"queued", "in_progress", "running", "pending", "searching"}
_TERMINAL_STATES = {"completed", "failed", "cancelled"}
_SUPPORTED_EVENT_KINDS = {
    "exec",
    "tool",
    "web",
    "search",
    "diff",
    "image",
    "approval",
    "collab",
    "plan",
}
_COMPLETION_CHUNK_OBJECTS = {
    "chat.completion.chunk",
    "chat.completions.chunk",
}
_COMPLETION_OBJECTS = {
    "chat.completion",
    "chat.completions",
}


def _to_text(*, value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, list):
        return "".join(_to_text(value=item) for item in value)
    if isinstance(value, dict):
        text = value.get("text")
        if isinstance(text, str):
            return text

        for key in ("content", "value", "message"):
            if key in value:
                nested = _to_text(value=value.get(key))
                if nested != "":
                    return nested

    return ""


def _event_object(*, event: dict) -> Optional[str]:
    obj = event.get("object")
    if isinstance(obj, str) and obj.strip() != "":
        return obj.strip().lower()

    evt_type = event.get("type")
    if isinstance(evt_type, str) and evt_type.strip() != "":
        return evt_type.strip().lower()

    return None


def _choices(*, event: dict) -> list[dict]:
    raw_choices = event.get("choices")
    if not isinstance(raw_choices, list):
        return []

    parsed: list[dict] = []
    for raw in raw_choices:
        if isinstance(raw, dict):
            parsed.append(raw)

    return parsed


def _is_completion_chunk_event(*, event: dict) -> bool:
    obj = _event_object(event=event)
    if obj in _COMPLETION_CHUNK_OBJECTS:
        return True
    if obj in _COMPLETION_OBJECTS:
        return False

    for choice in _choices(event=event):
        if isinstance(choice.get("delta"), dict):
            return True
    return False


def _is_completion_event(*, event: dict) -> bool:
    obj = _event_object(event=event)
    if obj in _COMPLETION_OBJECTS:
        return True
    if obj in _COMPLETION_CHUNK_OBJECTS:
        return False

    for choice in _choices(event=event):
        if isinstance(choice.get("message"), dict):
            return True
    return False


def _choice_index(*, choice: dict) -> int:
    index = choice.get("index")
    if isinstance(index, int) and index >= 0:
        return index
    return 0


def _choice_delta_text(*, choice: dict) -> str:
    delta = choice.get("delta")
    if not isinstance(delta, dict):
        return ""

    content = delta.get("content")
    return _to_text(value=content)


def _choice_message_text(*, choice: dict) -> str:
    message = choice.get("message")
    if isinstance(message, dict):
        content = message.get("content")
        text = _to_text(value=content)
        if text != "":
            return text

    text = choice.get("text")
    if isinstance(text, str):
        return text

    return ""


def _choice_finish_reason(*, choice: dict) -> Optional[str]:
    reason = choice.get("finish_reason")
    if not isinstance(reason, str):
        return None

    normalized = reason.strip().lower()
    if normalized in ("", "none", "null"):
        return None

    return normalized


class CompletionsThreadAdapter(ThreadAdapter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._active_events_by_key: dict[str, Element] = {}

    async def stop(self) -> None:
        await super().stop()
        self._active_events_by_key.clear()

    async def handle_custom_event(
        self,
        *,
        messages: Element,
        event: dict,
    ) -> None:
        event_type = event.get("type")
        if event_type not in ("agent.event", "codex.event"):
            return

        source = event.get("source")
        if not isinstance(source, str) or source.strip() == "":
            source = "agent" if event_type == "agent.event" else "codex"
        source = source.strip()

        name = event.get("name")
        if not isinstance(name, str) or name.strip() == "":
            name = event.get("event_type")
        if not isinstance(name, str) or name.strip() == "":
            name = event_type
        name = name.strip()

        kind = event.get("kind")
        if not isinstance(kind, str) or kind.strip() == "":
            return
        kind = kind.strip().lower()
        if kind not in _SUPPORTED_EVENT_KINDS:
            return

        state = event.get("state")
        if not isinstance(state, str) or state.strip() == "":
            state = "info"
        state = state.strip().lower()

        method = event.get("method")
        if not isinstance(method, str) or method.strip() == "":
            method = name
        method = method.strip()

        summary = event.get("summary")
        if not isinstance(summary, str) or summary.strip() == "":
            summary = method
        summary = summary.strip()

        headline = event.get("headline")
        if not isinstance(headline, str):
            headline = ""
        headline = headline.strip()

        item_id = event.get("item_id")
        if not isinstance(item_id, str):
            item_id = ""

        item_type = event.get("item_type")
        if not isinstance(item_type, str):
            item_type = ""
        event_path = event.get("path")
        if not isinstance(event_path, str):
            event_path = ""
        preview = event.get("preview")
        if not isinstance(preview, str):
            preview = ""

        raw_details = event.get("details")
        details: str
        if isinstance(raw_details, list):
            detail_lines = [
                line.strip() for line in raw_details if isinstance(line, str)
            ]
            details = "\n".join(line for line in detail_lines if line != "")
        elif isinstance(raw_details, str):
            details = raw_details.strip()
        else:
            details = ""

        data = event.get("data")
        if not isinstance(data, str):
            data = json.dumps(event, ensure_ascii=False, default=str)
        persisted_data = data if kind == "diff" and data != "" else ""

        correlation_key = event.get("correlation_key")
        if not isinstance(correlation_key, str) or correlation_key.strip() == "":
            correlation_key = event.get("event_key")
        if not isinstance(correlation_key, str) or correlation_key.strip() == "":
            correlation_key = None
        else:
            correlation_key = correlation_key.strip()

        in_progress = state in _ACTIVE_STATES
        now = datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")

        event_element: Element | None = None
        if correlation_key is not None:
            event_element = self._active_events_by_key.get(correlation_key)

        if event_element is None:
            attributes = {
                "id": str(uuid.uuid4()),
                "source": source,
                "name": name,
                "kind": kind,
                "state": state,
                "method": method,
                "item_id": item_id,
                "item_type": item_type,
                "path": event_path,
                "preview": preview,
                "summary": summary,
                "headline": headline,
                "details": details,
                "created_at": now,
                "updated_at": now,
            }
            if persisted_data != "":
                attributes["data"] = persisted_data
            event_element = messages.append_child(
                tag_name="event",
                attributes=attributes,
            )
        else:
            event_element.set_attribute("source", source)
            event_element.set_attribute("name", name)
            event_element.set_attribute("kind", kind)
            event_element.set_attribute("state", state)
            event_element.set_attribute("method", method)
            event_element.set_attribute("item_id", item_id)
            event_element.set_attribute("item_type", item_type)
            event_element.set_attribute("path", event_path)
            if preview != "" or event_element.get_attribute("preview") in (None, ""):
                event_element.set_attribute("preview", preview)
            event_element.set_attribute("summary", summary)
            event_element.set_attribute("headline", headline)
            if details != "" or event_element.get_attribute("details") in (None, ""):
                event_element.set_attribute("details", details)
            if persisted_data != "":
                event_element.set_attribute("data", persisted_data)
            event_element.set_attribute("updated_at", now)

        if correlation_key is not None:
            if in_progress:
                self._active_events_by_key[correlation_key] = event_element
            elif state in _TERMINAL_STATES:
                self._active_events_by_key.pop(correlation_key, None)

    async def _process_llm_events(self) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        messages = self._thread.root.get_children_by_tag_name("messages")
        if len(messages) == 0:
            raise RoomException("messages element is missing from thread document")
        doc_messages = messages[0]

        updates: asyncio.Queue = asyncio.Queue()
        update_thread_stop = object()
        message_elements_by_choice: dict[int, Element] = {}
        partial_text_by_choice: dict[int, str] = {}

        local_participant_name = self._room.local_participant.get_attribute("name")
        if not isinstance(local_participant_name, str):
            local_participant_name = ""

        async def update_thread() -> None:
            while True:
                entry = await updates.get()
                if entry is update_thread_stop:
                    break

                element, text = entry
                changes: dict[Element, str] = {element: text}
                should_stop = False
                while True:
                    try:
                        pending = updates.get_nowait()
                    except asyncio.QueueEmpty:
                        break

                    if pending is update_thread_stop:
                        should_stop = True
                        break

                    pending_element, pending_text = pending
                    changes[pending_element] = pending_text

                for changed_element, changed_text in changes.items():
                    changed_element["text"] = changed_text

                if should_stop:
                    break

        def ensure_message_element(*, choice_index: int) -> Element:
            existing = message_elements_by_choice.get(choice_index)
            if existing is not None:
                return existing

            element = doc_messages.append_child(
                tag_name="message",
                attributes={
                    "text": "",
                    "created_at": datetime.now(timezone.utc)
                    .isoformat()
                    .replace("+00:00", "Z"),
                    "author_name": local_participant_name,
                },
            )
            message_elements_by_choice[choice_index] = element
            return element

        def queue_text_update(*, choice_index: int, text: str) -> None:
            element = ensure_message_element(choice_index=choice_index)
            updates.put_nowait((element, text))

        def set_final_text(*, choice_index: int, text: str) -> None:
            queue_text_update(choice_index=choice_index, text=text)
            with tracer.start_as_current_span("chatbot.thread.message") as span:
                span.set_attribute("from_participant_name", local_participant_name)
                span.set_attribute("role", "assistant")
                span.set_attribute("text", text)

        update_thread_task = asyncio.create_task(update_thread())
        try:
            while True:
                evt = await self._llm_messages.get()
                if not isinstance(evt, dict):
                    continue

                if _is_completion_chunk_event(event=evt):
                    for choice in _choices(event=evt):
                        choice_index = _choice_index(choice=choice)

                        delta = _choice_delta_text(choice=choice)
                        if delta != "":
                            current = partial_text_by_choice.get(choice_index, "")
                            next_text = current + delta
                            partial_text_by_choice[choice_index] = next_text
                            queue_text_update(choice_index=choice_index, text=next_text)

                        if _choice_finish_reason(choice=choice) is not None:
                            final_text = partial_text_by_choice.get(choice_index, "")
                            if (
                                final_text != ""
                                or choice_index in message_elements_by_choice
                            ):
                                set_final_text(
                                    choice_index=choice_index,
                                    text=final_text,
                                )

                            partial_text_by_choice.pop(choice_index, None)
                    continue

                if _is_completion_event(event=evt):
                    choices = _choices(event=evt)
                    if len(choices) == 0:
                        text = _to_text(value=evt.get("text"))
                        if text != "":
                            set_final_text(choice_index=0, text=text)
                        continue

                    for choice in choices:
                        choice_index = _choice_index(choice=choice)
                        text = _choice_message_text(choice=choice)
                        if text == "":
                            text = partial_text_by_choice.get(choice_index, "")

                        if text == "":
                            continue

                        partial_text_by_choice[choice_index] = text
                        set_final_text(choice_index=choice_index, text=text)
                        partial_text_by_choice.pop(choice_index, None)
                    continue

                await self.handle_custom_event(messages=doc_messages, event=evt)
        except asyncio.QueueShutDown:
            pass
        finally:
            updates.put_nowait(update_thread_stop)

        await update_thread_task
